//Done by Vidal Alcázar Saiz for the ORTS competition held in 2007
//Based on the sample code by Michael Buro and Tim Furtak

#include "PlannerEventHandler.H"

#include "GameStateModule.H"
#include "GfxModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameTile.H"
#include "ServerObjData.H"
#include "Options.H"
#include "GUI.H"

#include "DrawOnTerrain2D.H"
#include "TerrainModule.H"
#include "TerrainBase.H"

#include <sys/time.h>


using namespace std;

enum { TILE_UNKNOWN=0, TILE_WATER=1, TILE_GROUND=2, TILE_CLIFF=3 };

bool PlannerEventHandler::handle_event(const Event& e){

  if (e.get_who() == GameStateModule::FROM) {

    if (e.get_what() == GameStateModule::STOP_MSG) {
      state.quit = true; return true;
    }
    
    if (e.get_what() == GameStateModule::READ_ERROR_MSG) {
      state.quit = state.error = true; return true;
    }
    
    if (e.get_what() == GameStateModule::VIEW_MSG) {

	if(time_behind < 1.0){
		double time = secs();
	    compute_actions();
		time_behind += secs()-time;
	}else
		time_behind -= 0.125;

	state.gsm->send_actions();

      if (state.gui) {
        state.gui->event();
        if (state.gui->quit) exit(0);
      }

      // as a hack for flickering, gui->display() is called in dot.start()
      // so call everything that draws to the debug window between start() and end()
      DrawOnTerrain2D dot(state.gui);
      dot.start();
      dot.draw_line(-60, 10, -20, 10, Vec3<real4>(1.0, 1.0, 1.0));
      dot.draw_line(-60, 10, -65, 15, Vec3<real4>(1.0, 1.0, 1.0));
      dot.draw_line(-44, 10, -56, 40, Vec3<real4>(1.0, 1.0, 1.0));
      dot.draw_line(-28, 10, -36, 40, Vec3<real4>(1.0, 1.0, 1.0));
      dot.draw_line(-36, 40, -30, 34, Vec3<real4>(1.0, 1.0, 1.0));

      dot.draw_circle(-40, 25, 35, Vec3<real4>(1.0, 0.0, 1.0));

      dot.end();

      return true;
    }
  }
    
  return false;
}

//returns current times in seconds with microseconds precision
double PlannerEventHandler::secs(){
	timeval tim;
	gettimeofday(&tim, NULL);
	return tim.tv_sec+(tim.tv_usec/1000000.0);
}



//taken from Distance (slightly modified)
real8 distance_worker_mineral(sint4 wx, sint4 wy, sint4 mx, sint4 my){// return 1; }
  ScalarPoint delta(wx - mx, wy - my);
  Scalar r = FixedPoint::add(WORKER_RADIUS, MINERAL_RADIUS);
  return delta.l2_norm() - r;
}

//taken from Distance (slightly modified)
real8 dist_c_p(sint4 cx, sint4 cy, sint4 r, sint4 px, sint4 py){
  return sqrt(square(real8(cx-px)) + square(real8(cy-py))) - r;
}

//taken from Distance (slightly modified)
real8 distance_cc_worker(const Object &e1, sint4 cx, sint4 cy){
  sint4 o2x1, o2y1, o2x2, o2y2;
  e1.get_p1(o2x1, o2y1);
  e1.get_p2(o2x2, o2y2);


  if (cx < o2x1) {
    
    if (cy < o2y1) return dist_c_p(cx, cy, WORKER_RADIUS, o2x1, o2y1);
    if (cy > o2y2) return dist_c_p(cx, cy, WORKER_RADIUS, o2x1, o2y2);
    return dist_c_p(cx, cy, WORKER_RADIUS, o2x1, cy);
    
  }
  if (cx > o2x2) {
    
    if (cy < o2y1) return dist_c_p(cx, cy, WORKER_RADIUS, o2x2, o2y1);
    if (cy > o2y2) return dist_c_p(cx, cy, WORKER_RADIUS, o2x2, o2y2);
    return dist_c_p(cx, cy, WORKER_RADIUS, o2x2, cy);
    
  }
  
  if (cy < o2y1) return dist_c_p(cx, cy, WORKER_RADIUS, cx, o2y1);
  if (cy > o2y2) return dist_c_p(cx, cy, WORKER_RADIUS, cx, o2y2);
  return -1;
}

void PlannerEventHandler::initialize(){

	const Game &game = state.gsm->get_game();

	const Map<GameTile> &map = game.get_map();

	const Game::ObjCont &objs = game.get_objs(game.get_client_player());

	width = map.get_width()/COARSE_VALUE;
	height = map.get_height()/COARSE_VALUE;
	tile_points = (COARSE_VALUE*game.get_tile_points());

	//initialization of the values of the base's variables
	FORALL (objs, it) {
		 GameObj *gob = (*it)->get_GameObj();
		const ServerObjData &sod = gob->sod;
		if (!gob || !sod.in_game || *sod.x < 0 || gob->is_dead()) { continue; }

		if(gob->bp_name().compare("controlCenter") == 0){
			controlX_precise = *gob->sod.x;
			controlY_precise = *gob->sod.y;
			controlX = (int)(controlX_precise/tile_points);
			controlY = (int)(controlY_precise/tile_points);
			cc = gob;
			break;
		}
	}


	terrain_cost = new int*[width];
	FORS(i, width) terrain_cost[i] = new int[height];
	FORS(i, width) FORS(j, height) terrain_cost[i][j] = 0;

	//initialization of the costs of the map due to static obstacles
	FORS(i,map.get_width()){
		FORS(j,map.get_height()){
			const GameTile &tile = map(i,j);
			if(tile.get_min_h() > 0)
				terrain_cost[(int)(floor(i/COARSE_VALUE))][(int)(floor(j/COARSE_VALUE))] += ADDITIONAL_HIGHTILE_COST;
		}
	}

	last_action = new double[NUMBER_OF_WORKERS];	(double*) calloc(NUMBER_OF_WORKERS, sizeof(double));
	FORS(i, NUMBER_OF_WORKERS) last_action[i] = 0;

	goals = new TerrainBase::Loc[NUMBER_OF_WORKERS];	(TerrainBase::Loc*) malloc(NUMBER_OF_WORKERS * sizeof(TerrainBase::Loc));
	FORS(i,NUMBER_OF_WORKERS) goals[i].x = -1;

	time_behind = 0;

	terrain_cost_iteration = new int*[width];
	FORS(i, width) terrain_cost_iteration[i] = new int[height];

	minerals = new int*[width];
	FORS(i, width) minerals[i] = new int[height];

	idle = new bool[NUMBER_OF_WORKERS];

}


//calculates a coarse cost from src to dest
int calculate_path_cost(int **terrain_cost_iteration, int destX, int destY, int srcX, int srcY, int width, int height){
	if(destX < 0 || destX >= width || destY < 0 || destX >= height) return 500;
	int cost = 0, inc = 1;
	if(destX == srcX){ //vertical slope
		if(destY<srcY) inc = -1;
		for(int i = srcY; i!=destY+inc;i+=inc)
			cost += terrain_cost_iteration[srcX][i];
	}else{
		const float slope = ((float)destY - (float)srcY) / ((float)destX - (float)srcX);
		float incY = 1;
		if(destX<srcX) inc = -1;
		if(destY<srcY) incY = -1;
		float stepY = srcY;
		for(int i = srcX; i!=destX+inc;i+=inc,stepY+=(slope*inc))
			for(int j = (int)(stepY); j!=(int)(stepY+(slope*inc))+incY && j!=destY+incY;j+=incY)
				cost += terrain_cost_iteration[i][j];
	}
	return cost;
}

//adds extra cost to the coarse tiles between src and dest
void increase_cost(int **terrain_cost_iteration, sint4 destX, sint4 destY, int srcX, int srcY, int width, int height){
	if(destX < 0 || destX >= width || destY < 0 || destX >= height) return;
	int inc = 1;
	if(destX == srcX){ //vertical slope
		if(destY<srcY) inc = -1;
		for(int i = srcY; i!=destY+inc;i+=inc)
			terrain_cost_iteration[srcX][i] += ADDITIONAL_WORKER_COST;
	}else{
		const float slope = ((float)destY - (float)srcY) / ((float)destX - (float)srcX);
		float incY = 1;
		if(destX<srcX) inc = -1;
		if(destY<srcY) incY = -1;
		float stepY = srcY;
		for(int i = srcX; i!=destX+inc;i+=inc,stepY+=(slope*inc))
			for(int j = (int)(stepY); j!=(int)(stepY+(slope*inc))+incY && j!=destY+incY;j+=incY)
				terrain_cost_iteration[i][j] += ADDITIONAL_WORKER_COST;
	}
}


void PlannerEventHandler::compute_actions(){

	//initialization of the global variables
	if(!initialized){ initialize(); initialized = true;}

	const Game &game = state.gsm->get_game();

	const Game::ObjCont &objs = game.get_objs(game.get_client_player());

	FORS(i, width) FORS(j, height) terrain_cost_iteration[i][j] = terrain_cost[i][j];

	FORS(i, width) FORS(j, height) minerals[i][j] = 0;

	const Game::ObjCont &n_objs = game.get_objs(1);

	//initialization of the position of the minerals and calculation of the costs added by sheeps
	FORALL (n_objs, it) {
		const GameObj *gob = (*it)->get_GameObj();
		const ServerObjData &sod = gob->sod;
		if (!gob || !sod.in_game || *sod.x < 0 || gob->is_dead()) { continue; }
		if(gob->bp_name().compare("mineral") == 0)
			minerals[(int)floor(*gob->sod.x/tile_points)][(int)floor(*gob->sod.y/tile_points)]++;
		else if(gob->bp_name().compare("sheep") == 0)
			terrain_cost_iteration[(int)floor(*gob->sod.x/tile_points)][(int)floor(*gob->sod.y/tile_points)]+=ADDITIONAL_SHEEP_COST;
	}

	int index = 0;
	int sim_workers = 0;
	double currentTime = secs();

	FORS(i, NUMBER_OF_WORKERS) idle[i] = 0;

	if(state.gsm->get_game().get_action_frame() >1)
	FORALL (objs, it){
		if(sim_workers==MAX_PATH) { break; }	//maybe some of the workers will never have the chance to act

		GameObj *gob = (*it)->get_GameObj();
		const ServerObjData &sod = gob->sod;
		if(!gob || !sod.in_game || *sod.x < 0 || gob->is_dead()) { continue; }
		if(!gob->is_pending_action() && !*sod.is_moving && !*sod.speed && last_action[index] + PATHFINDING_TIME < currentTime){
			last_action[index] = currentTime;
			idle[index] = true;
			sim_workers++;
		}
		index++;
	}

	index = 0;
	FORALL (objs, it) {

		GameObj *gob = (*it)->get_GameObj();
		const ServerObjData &sod = gob->sod;

		if (!gob || !sod.in_game || *sod.x < 0 || gob->is_dead()) { continue; }

		if (gob->bp_name().compare("worker") != 0)  { continue; }

		int workerX = (int)(*gob->sod.x/tile_points);
		int workerY = (int)(*gob->sod.y/tile_points);

		//case 1: goes to mine
		if(gob->get_int("minerals") < 10){

			//case 1a: not in a tile with minerals (coarse pathfinding)
			if(minerals[workerX][workerY] == 0){ // && last_action[index] + PATHFINDING_TIME < currentTime){
				if(idle[index]){
					ofstream problem;
					problem.open("/tmp/workers_problem.pddl");

					problem << "(define (problem workers-problem) (:domain coarse-worker) (:objects " << endl;

					//positions declaration
					FORS(i,width)
						FORS(j,height)
							problem << "pos" << i << "-" << j << " - position " << endl;

					//worker declaration
					problem << "obj" << game.get_cplayer_info().get_id(gob) << " - worker" << endl;

					//mineral declaration
					FORS(i,width)
						FORS(j,height)
							if(minerals[i][j] > 0)	problem << "min" << i << "-" << j << " - goalobj " << endl;

					problem << ") (:init " << endl;

					//positions instantiation
					FORS(i,width){
						FORS(j,height){
							problem << "(= (posX pos" << i << "-" << j << ") " << i << ")" << endl;
							problem << "(= (posY pos" << i << "-" << j << ") " << j << ")" << endl;
							problem << "(= (positionCost pos" << i << "-" << j << ") " << calculate_path_cost(terrain_cost_iteration, i, j, workerX, workerY, width, height) << ")" << endl;
						}
					}

					//metric instantiation
					problem << "(= (total-cost) 0)" << endl;

					//worker instantiation
					problem << "(at obj" << game.get_cplayer_info().get_id(gob) << " pos" << workerX << "-" << workerY << ")" << endl;

					//mineral instantiation
					FORS(i,width)
						FORS(j,height)
							if(minerals[i][j] > 0)	problem << "(at min" << i << "-" << j << " pos" << i << "-" << j << ")" << endl;

	 				//goal
					problem << ") (:goal (atgoal obj" << game.get_cplayer_info().get_id(gob) << ") )" << endl;

					//metric
					problem << "(:metric minimize (total-cost) ) )" << endl;

					problem.close();

					//call to planner
					system("~rts13s/ff -o ~rts13s/coarse-worker.pddl -f /tmp/workers_problem.pddl -g 1 -h 1 -O | sed -n -e '/0:/,/^$/p' | sed -e 's/.*:[ \t]*//' | sed -e 's/OBJ//g' | sed -e 's/POS//g' | sed -e 's/-/ /g' > /tmp/output.txt");

					//read output
					string line;
					ifstream infile ("/tmp/output.txt");
					if (infile.is_open()){
						getline (infile,line);

						char c_line[line.size()];
						strcpy(c_line, line.c_str());

						char* action = strtok(c_line," \n\t");
						char* obj_str = strtok(NULL," \n\t");
	 					char* x_str = strtok(NULL," \n\t");
						char* y_str = strtok(NULL," \n\t");

						if(strcmp(action, "MOVE") == 0){
								TerrainBase::Loc target = TerrainBase::Loc((sint4)((atoi(x_str))*tile_points + tile_points/2),(sint4)((atoi(y_str))*tile_points + tile_points/2));
								call_handlers(TerrainModule::PathEvent(EventFactory::new_who(), TerrainModule::FIND_PATH_MSG,TerrainBase::Obj2LocTask(gob, target)));
								goals[index].x = atoi(x_str); goals[index].y = atoi(y_str);
								increase_cost(terrain_cost_iteration, atoi(x_str), atoi(y_str), workerX, workerY, width, height);
						}
						infile.close();
					}//end open

				}else if(goals[index].x != -1  && !*sod.is_moving && !*sod.speed){
					Vector<sint4> params;
					params.push_back(goals[index].x*tile_points + tile_points/2);
					params.push_back(goals[index].y*tile_points + tile_points/2);
					params.push_back(*gob->sod.max_speed);
					gob->set_action("move", params);

					increase_cost(terrain_cost_iteration, goals[index].x, goals[index].y, workerX, workerY, width, height);
				}else{
					terrain_cost_iteration[workerX][workerY] += ADDITIONAL_WORKER_COST;
				}
			}else{	//case 1b: in a tile with minerals (fine pathfinding)
					terrain_cost_iteration[workerX][workerY] += ADDITIONAL_WORKER_COST;

					GameObj *min_obj;

					real8 distance = 1000;

					FORALL (n_objs, it) {
						GameObj *n_gob = (*it)->get_GameObj();
						const ServerObjData &n_sod = n_gob->sod;
						if (!n_gob || !n_sod.in_game || *n_sod.x < 0 || n_gob->is_dead()) { continue; }
						if(n_gob->bp_name().compare("mineral") == 0){
							real8 distance_aux = distance_worker_mineral(*sod.x, *sod.y, *n_sod.x, *n_sod.y);
							if(distance > distance_aux){
								min_obj = n_gob;
								distance = distance_aux;
							}
						}
					}

					if(distance > MINERAL_SIZE){
						if(!idle[index] && !*sod.is_moving && !*sod.speed){
							Vector<sint4> params;
							params.push_back(*min_obj->sod.x);
							params.push_back(*min_obj->sod.y);
							params.push_back(*gob->sod.max_speed);
							gob->set_action("move", params);
						}else{
							call_handlers(TerrainModule::PathEvent(EventFactory::new_who(), TerrainModule::FIND_PATH_MSG,TerrainBase::Obj2ObjTask(gob, min_obj)));
							}
					}else if(!gob->component("pickaxe")->get_int("active")){
						Vector<sint4> args(1, game.get_cplayer_info().get_id(min_obj));
						gob->component("pickaxe")->set_action("mine", args);
					}
			}
		}else{	//case 2: goes to deploy
			goals[index].x = controlX;
			goals[index].y = controlY;

			if(distance_cc_worker(*cc, *sod.x, *sod.y) > CC_SIZE){
				increase_cost(terrain_cost_iteration, controlX, controlY, workerX, workerY, width, height);
				if(!idle[index] && !*sod.is_moving && !*sod.speed){
					Vector<sint4> params;
					params.push_back(controlX_precise);
					params.push_back(controlY_precise);
					params.push_back(*gob->sod.max_speed);
        			gob->set_action("move", params);
				}else
					call_handlers(TerrainModule::PathEvent(EventFactory::new_who(), TerrainModule::FIND_PATH_MSG,TerrainBase::Obj2ObjTask(gob, cc)));
			}else{
				Vector<sint4> args;
				args.push_back(game.get_cplayer_info().get_id(cc));
				gob->set_action("return_resources", args);
				PlayerInfo &pi = game.get_cplayer_info();
				ScriptObj *player = pi.global_obj("player");
				cout << "Mineral gathered: " << player->get_int("minerals")+10 << endl;
			}

		}//end cases

		index++;

	}//end loop workers

}//end compute actions

