//Done by Vidal Alcázar Saiz for the ORTS competition held in 2007
//Based on the sample code by Michael Buro and Tim Furtak

#include "Global.H"
#include "Options.H"
#include "GameStateModule.H"
#include "GfxModule.H"
#include "ShutDown.H"
#include "SDLinit.H"
#include "Game.H"
#include "GUI.H"

#include "PlannerEventHandler.H"
#include "ClientState.H"
#include "Terrain.H"
#include "TerrainModule.H"

#include "SimplePathfinder.H"
#include "ForceFieldPE.H"
#include "SimplePFAdapter.H"


using namespace std;

//===================================================================

static char *VERSION = "Plannerai v0.1";

static const sint4 GFX_X = 900;  // graphics window size
static const sint4 GFX_Y = 740;

//===================================================================

void add_options()
{
  Options opt("PlannerClient");
  opt.put("-disp",  "2d display");
}

//===================================================================

// loop is called

class Looper : public LoopFunctor
{
public :

  Looper(PlannerAIState *state_) : state(state_) {}
  
  void loop() const {
    if (!state->gsm->recv_view()) SDL_Delay(1);
  }

private:
  PlannerAIState *state;
};

//===================================================================

int main(int argc, char **argv)
{
  signals_shut_down(true);  
  
  // populate opt with options from various modules

  add_options();
  GameStateModule::Options::add();
  Game::Options::add();
  GUI::add_options();

  // process command line arguments
  if (Options::process(argc, argv, cerr, VERSION)) exit(20);

  // initialize SDL
  SDLinit::video_init();
  SDLinit::network_init();
  
  // populate option objects with command line values
  // and create modules
  
  GameStateModule::Options gsmo;
  GameStateModule gsm(gsmo);


  // change seed value to time(0) if 0

  sint4 seed;
  Options::get("-seed", seed);
  if (!seed) seed = time(0);
  Options::set("-seed", seed);
  cout << "seed=" << seed << endl;

  // debug graphics

  GUI *gui = 0;
  bool disp;
  Options::get("-disp", disp);
  if (disp) {
    SDLinit::video_init();
    gui = new GUI;
  }


  // gives event handler access to the state
  PlannerAIState state(&gsm, gui);

  PlannerEventHandler sev(state, seed);

  // gsm events are handled in sev
  gsm.add_handler(&sev);

  ForceFieldPE pe;
  SimplePFAdapter gp;
  SimplePathfinder timp(gp,pe);

  //Terrain timp;
  TerrainModule tm(gsm, timp, 80);
  sev.add_handler(&tm);

  // connect game state module to server
  if (!gsm.connect()) ERR("connection problems");

  if (gui) {
    gui->init(GFX_X, GFX_Y, gsm.get_game());
    gui->display();
  }

  Looper l(&state);

  while (!state.quit)l .loop();

  return 0;
}
