// $Id: Net_Invitation.C,v 1.1 2005/08/23 21:24:53 hakman Exp $

// (c) Michael Buro 2003, mburo@cs.ualberta.ca, licensed under the GPL

#include  "Net_Invitation.H"
using namespace std;

Net_Invitation::Net_Invitation(Invite invite)
{

  iid           = 0;
  owner         = invite.get_sender();
  name          = invite.get_game_name();
  maxPlayers    = invite.get_nplayers();

  invisible     = invite.get_invisible();
  anony_ids     = invite.get_anony_ids();
  alias         = invite.get_alias();
  passprotected = invite.get_passprotected();
  password      = invite.get_password();


  cout << "===========================================" << endl;
  cout << password << endl;
  cout << "===========================================" << endl;
  stringstream ss;

  ss << "-nplayers ";
  ss << maxPlayers;
  ss << " ";

  options     = ss.str();
  options     += invite.get_options();

  playerCount = 0;
  readyCount  = 0;

  updateInfo();
}

Net_Invitation::Net_Invitation(const Net_Invitation& n)
{

  owner       = n.owner;
  name        = n.name;
  maxPlayers  = n.maxPlayers;
  invisible   = n.invisible;
  anony_ids   = n.anony_ids;
  password    = n.password;

  options     = n.options;

  playerCount = n.playerCount;
  readyCount  = n.readyCount;

  info = n.info;

  players = n.players;
  present = n.present; 
  ready   = n.ready;
  
}

Net_Invitation::~Net_Invitation() { }

void Net_Invitation::setIID(uint4 iid_) {
  
  iid = iid_;
  updateInfo();
  
}


void Net_Invitation::updateInfo()
{

  info = Invit_Disc::compose(iid,
                             playerCount,
                             owner,
                             name,
                             maxPlayers,
                             invisible,
                             anony_ids,
			     passprotected,
			     options);

  cout << "Net_Invitation -> unpdateinfo: invit_disc: " << info << endl;
}


sint4 Net_Invitation::addPlayer(const string &name_, const string &alias_)
{
  if (playerCount == maxPlayers) return -1;
  if (getPlayer(name_) >= 0) return -1;

  /** find an empty spot */
  for(uint4 i = 0; i < players.size(); i++){
    if(!present[i]){
      aliases[i] = alias_;
      players[i] = name_;
      ready[i] = false;
      present[i] = true;
      playerCount++;
      updateInfo();
      return i;
    }
  }
  
  aliases.push_back(alias_);
  players.push_back(name_);
  ready.push_back(false);
  present.push_back(true);
  playerCount ++;
  updateInfo();
  return (players.size()-1);
}

sint4 Net_Invitation::removePlayer(const string &name_)
{

  int index = getPlayer(name_);

  if(index >= 0){
      present[index] = false;
      ready[index]   = false;
      playerCount--;
      updateInfo();
  }
  

  return index;
}

sint4 Net_Invitation::removePlayer(sint4 pid)
{
  if (pid >= (sint4)maxPlayers || pid < 0) return playerCount;

  present[pid] = false;
  ready[pid]   = false;
  playerCount--;
  updateInfo();
  return pid;
}

sint4 Net_Invitation::getPlayer(const string& name_)
{
  for (uint4 i = 0; i < players.size(); i++) {
    if(present[i] && players[i] == name_) return i;
  }
  return -1;
}

int Net_Invitation::setReady(const string& name, bool ready_)
{
  int i = getPlayer(name);

  if (i < 0) return i;

  if(ready[i] == ready_) return i;

  if(ready_) { readyCount++; } else { readyCount--; }

  ready[i] = ready_;

  return i;
}


void Net_Invitation::print()
{
    cout << endl << "\t-----Game-----" << endl;
    cout << info << endl;
    for(uint4 i = 0; i < players.size(); i++){
      if(!present[i]){
        continue;
      }
      cout << "\t   ["<< i <<"] " << players[i];
      if(ready[i]){
        cout << " R";
      }else{
        cout << " U";
      }

      if(present[i]){
        cout << " P" << endl;
      }else{
        cout << " N" << endl;
      }
  }
}
  


