#ifndef LINALG_H
#define LINALG_H

#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include"GLee.h"
#include<GL/gl.h>

#ifndef number
#define number float
#endif

#define nrand() (((float)rand())/(RAND_MAX+1.0))

class vec3
{
  public:
    vec3(number X,number Y,number Z){ x=X;y=Y;z=Z; };
    vec3(const number *B){x=B[0];y=B[1];z=B[2];};
    vec3(){x=0;y=0;z=0;};

    union{
      struct{
	number x;
	number y;
	number z;
      };
      struct{
	number r;
	number g;
	number b;
      };
      number A[3];
    };

    operator const number *()const{return A;};
    operator number *(){return A;};
    const number & operator [](int i) const {return A[i];};
    number & operator [](int i){return A[i];};

    vec3 operator +(const vec3 &v) const
      {return vec3(x+v.x, y+v.y, z+v.z);};
    vec3 operator -(const vec3 &v) const
      {return vec3(x-v.x, y-v.y, z-v.z);};
    vec3 operator *(const vec3 &v) const
      {return vec3(x*v.x, y*v.y, z*v.z);};
    vec3 operator /(const vec3 &v) const
      {return vec3(x/v.x, y/v.y, z/v.z);};

    vec3 operator +(const number &n) const {return vec3(x+n, y+n, z+n);};
    vec3 operator -(const number &n) const {return *this+(-n);};
    vec3 operator *(const number &n) const {return vec3(x*n, y*n, z*n);};
    vec3 operator /(const number &n) const {return *this*(1.0/n);};

    number length() const {return sqrt(x*x+y*y+z*z);};
    void normalize() {number l=length();x/=l;y/=l;z/=l;};
    number dot(const vec3 &v) const {return v.x*x + v.y*y + v.z*z;};
    vec3 cross(const vec3 &v) const 
    {return vec3(y*v.z - z*v.y, z*v.x - x*v.z, x*v.y - y*v.x);};
    vec3 rotate(vec3 axis,number a)
    {
      number c=cos(a),s=sin(a);
      return (*this)*c+cross(axis)*s+axis*dot(axis)*(1-c);
    }

    void print(FILE *f=stdout) { fprintf(f,"[%f, %f , %f]\n",x,y,z); };
    void write(FILE *f) { fwrite(A,12,1,f); };
    void read(FILE *f) { fread(A,12,1,f); };

    static vec3 vrand(){return vec3(2*nrand()-1,2*nrand()-1,2*nrand()-1);};
};

class vec4:public vec3
{
  public:
    vec4(number X,number Y,number Z,number W):vec3(X,Y,Z){w=W;};
    vec4(const number *B):vec3(B){w=B[3];};
    vec4(const vec3 &v,number W){x=v.x;y=v.y;z=v.z;w=W;};
    vec4():vec3(){w=1;};
    
    union{
      struct{
	number w;
      };
      struct{
	number a;
      };
    };
    
    vec4 operator +(const vec4 &v) const
      {return vec4(x+v.x, y+v.y, z+v.z, w+v.w);};
    vec4 operator -(const vec4 &v) const
      {return vec4(x-v.x, y-v.y, z-v.z, w-v.w);};
    vec4 operator *(const vec4 &v) const
      {return vec4(x*v.x, y*v.y, z*v.z, w*v.w);};
    vec4 operator /(const vec4 &v) const
      {return vec4(x/v.x, y/v.y, z/v.z, w/v.w);};

    vec4 operator +(const number &n) const {return vec4(x+n, y+n, z+n, w+n);};
    vec4 operator -(const number &n) const {return *this+(-n);};
    vec4 operator *(const number &n) const {return vec4(x*n, y*n, z*n, w*n);};
    vec4 operator /(const number &n) const {return *this*(1.0/n);};

    vec3 project() const {return (*this)/w;};
    vec4 cross(const vec4 &v) const {return vec4(vec3::cross(v),0);};
    number dot(const vec4 &v) const {return v.x*x + v.y*y + v.z*z + v.w*w;};

    void print(FILE *f=stdout) { fprintf(f,"[%f, %f , %f, %f]\n",x,y,z,w); };
    void write(FILE *f) { fwrite(A,16,1,f); };
    void read(FILE *f) { fread(A,16,1,f); };
};

class mat4
{
  public:
    mat4(){};
    mat4(vec4 r0,vec4 r1,vec4 r2,vec4 r3){A[0]=r0;A[1]=r1;A[2]=r2;A[3]=r3;};
    
    operator const number *()const{return &(A[0][0]);};
    operator number *(){return &(A[0][0]);};
    vec4 & operator [](int i){return A[i];};
    const vec4 & operator [](int i)const{return A[i];};

    vec4 col(int i)const{return vec4(A[0][i],A[1][i],A[2][i],A[3][i]);};
    void glGetProjection(){glGetFloatv(GL_TRANSPOSE_PROJECTION_MATRIX,*this);};
    void glGetModelview(){glGetFloatv(GL_TRANSPOSE_MODELVIEW_MATRIX,*this);};

    mat4 operator *(const mat4 &c) const;
    vec4 operator *(const vec4 &v) const;
    vec3 operator *(const vec3 &v) const;
    mat4 inv();
    mat4 trans();
    void identity();

    void print(FILE *f=stdout) { for(int i=0;i<4;i++)A[i].print(f); };
    void write(FILE *f) { fwrite(A,64,1,f); };
    void read(FILE *f) { fread(A,64,1,f); };

  private:
    vec4 A[4];
};

#endif
