//Done by Vidal Alcázar Saiz for the ORTS competition held in 2007
//Based on the sample code by Michael Buro and Tim Furtak

#ifndef PlannerEventHandler_H
#define PlannerEventHandler_H

#include "EventHandler.H"
#include "Random.H"
#include "GameObj.H"

#include "ClientModule.H"
#include "TerrainBase.H"

class GameStateModule;
class GUI;

#define ADDITIONAL_WORKER_COST 10	//cost that a worker adds
#define ADDITIONAL_SHEEP_COST 5		//cost that a sheep adds
#define ADDITIONAL_HIGHTILE_COST 25	//cost that a high tile adds
#define COARSE_VALUE 4				//abstraction level of the map grid
#define MINERAL_SIZE 4				//distance to be in touch
#define CC_SIZE 3					//distance to be in touch
#define MINERAL_RADIUS 6			//mineral radius
#define WORKER_RADIUS 3				//worker radius
#define NUMBER_OF_WORKERS 20		//number of workers
#define PATHFINDING_TIME 5.0		//time to wait until plans are recalculated in seconds
#define MAX_PATH 1					//maximum paths that can be searched at the same time


struct PlannerAIState {

  GameStateModule *gsm;
  GUI             *gui;  // !=0 : debug gfx used
  bool just_drew;  // just drew a gfx frame
  bool quit;       // true => quit game
  bool error;      // true <=> error occurred
  
  PlannerAIState(GameStateModule *gsm_, GUI *gui_) :
    gsm(gsm_), gui(gui_)
  {
    just_drew = quit = error = false;
  }

};


class PlannerEventHandler : public EventHandler, public ClientModule {

public:
  
  PlannerAIState &state;

  PlannerEventHandler(PlannerAIState &state_, sint4 seed) : state(state_) {
    rand.srand(seed);
	initialized = 0;
  }

  bool handle_event(const Event &e);

private:

	Random rand;

	int width;
	int height;
	int tile_points;

	int** terrain_cost;				//array that holds the static costs of the terrain
	int** terrain_cost_iteration;	//array that holds the dynamic costs of the terrain
	int** minerals;					//array that holds the position of the minerals

	int controlX;
	int controlY;
	sint4 controlX_precise;
	sint4 controlY_precise;
	GameObj* cc;

	bool initialized;
	double time_behind;

	bool* idle;						//state of the worker
	TerrainBase::Loc* goals;		//last target a given worker had

	double* last_action;			//last time something was planned for a given worker 
	double secs();					//maybe I should think of another name for this function
	void initialize();				//initialization of static and initial values
	void compute_actions();

  
};

#endif
