// $Id: PolyGraph.C,v 1.14 2005/05/28 19:14:23 orts_mburo Exp $

#include"PolyGraph.H"

PolyGraph::~PolyGraph() {
  while (!all_nodes.empty()) {
    Graph_node* current = all_nodes.back();
    while (!current->neighbours.empty()) {
      current->neighbours.pop_back();
    }
    all_nodes.pop_back();
  }
}

void PolyGraph::clear_graph() {
    if (all_nodes.empty()) {
	  return;
    }

    FORALL(all_nodes,i) {
	  Graph_node* current = *i;
	  while(!current->neighbours.empty()) {
		current->neighbours.pop_back();
	  }
    }

    while(!all_nodes.empty()) {
	  Graph_node* current = all_nodes.back();
	  all_nodes.pop_back();
	  
	  delete current;
    }
}
void PolyGraph::build_graph(Decomp* decomposition) {
  count = 0;
    
  //mark them as not visited
  FORALL(decomposition->nodes,i) {
    Decomp::Node *n = *i;
    if (n->type!=Decomp::Node::R_NODE) continue;

    n->visited = false;
  }
    
  Vector<Graph_node*> pts;
    
  //start building the graph
  FORALL(decomposition->nodes,j) {
    Decomp::Node *n = *j;
	  
    if (n->type!=Decomp::Node::R_NODE) {
      continue;
    }

    // n->write();
    
    //current rectangle
    Decomp::Rect r0 = n->r;

    Graph_node* new_node = new Graph_node();

    new_node->index = all_nodes.size();

    new_node->p.x = (r0.x0 + r0.x1)/2;
    new_node->p.y = (r0.y0 + r0.y1)/2;
    new_node->n1 = n;
    new_node->n2 = n;
    
    pts.push_back(new_node);
    all_nodes.push_back(new_node);
    
    //north neighbour
#if 0
    if (n->nl[0].empty()) {
      Point2i p;
      p.x = (r0.x0 + r0.x1)/2;
      p.y = r0.y0;

      Graph_node* new_node = new Graph_node();

      new_node->index = all_nodes.size();

	  
      new_node->p.x = p.x;
      new_node->p.y = p.y;

      new_node->n1 = n;
      new_node->n2 = NULL;

      pts.push_back(new_node);

      all_nodes.push_back(new_node);
    }
#endif
    FORALL(n->nl[0],it0) {
      Decomp::Node *neighbour = *it0;

	
      if (neighbour->type!=Decomp::Node::R_NODE) {
	cout<<"??"<<endl;
	continue;
      }

      if (neighbour->data != n->data || n->data == 1 || neighbour->data == 1) {
	continue;
      }
	
      //handle it differently if the neighbour rectangle is visited
      if (neighbour->visited) {
	Graph_node *already_exist = get_node(n,neighbour);
	if (already_exist==NULL) {
	  ERR("should not be happening");
	}

	pts.push_back(already_exist);

	continue;
      }

      Decomp::Rect r1 = neighbour->r;

      Point2i p1,p2,p3,p4;

      p1.x = r0.x0;
      p1.y = r0.y1;
      p2.x = r0.x1;
      p2.y = r0.y1;

      p3.x = r1.x0;
      p3.y = r1.y0;
      p4.x = r1.x1;
      p4.y = r1.y0;
		    
      Point2i p = get_point(p1,p2,p3,p4);

      Graph_node* new_node = new Graph_node;

      new_node->p = p;
      new_node->n1 = n;
      new_node->n2 = neighbour;

      pts.push_back(new_node);

      //add it to the list if it is not added yet

      new_node->index = all_nodes.size();
      all_nodes.push_back(new_node);
    }
    //east neighbour
#if 0
    if (n->nl[1].empty()) {
      Point2i p;
      p.x = r0.x1;
      p.y = (r0.y0 + r0.y1)/2;

      Graph_node* new_node = new Graph_node();
      new_node->p.x = p.x;
      new_node->p.y = p.y;

      new_node->n1 = n;
      new_node->n2 = NULL;

      pts.push_back(new_node);

      new_node->index = all_nodes.size();
      all_nodes.push_back(new_node);
    }
#endif
    FORALL(n->nl[1],it1) {
      Decomp::Node *neighbour = *it1;

      if (neighbour->type!=Decomp::Node::R_NODE) {
	cout<<"??"<<endl;
	continue;
      }
      if (neighbour->data != n->data || n->data == 1 || neighbour->data == 1) {
	continue;
      }

      //handle it differently if the neighbour rectangle is visited
      if (neighbour->visited) {
	Graph_node *already_exist = get_node(n,neighbour);
	if (already_exist==NULL) {
	  ERR("should not be happening");
	}

	pts.push_back(already_exist);

	continue;
      }
		
      Decomp::Rect r1 = neighbour->r;

      Point2i p1,p2,p3,p4;

      p1.x = r0.x1;
      p1.y = r0.y1;
      p2.x = r0.x1;
      p2.y = r0.y0;

      p3.x = r1.x0;
      p3.y = r1.y1;
      p4.x = r1.x0;
      p4.y = r1.y0;
		    
      Point2i p = get_point(p1,p2,p3,p4);

      Graph_node* new_node = new Graph_node;

      new_node->p = p;
      new_node->n1 = n;
      new_node->n2 = neighbour;

      pts.push_back(new_node);

      //add it to the list if it is not added yet
      new_node->index = all_nodes.size();
      all_nodes.push_back(new_node);
    }

    //south neighbour
#if 0
    if (n->nl[2].empty()) {
      Point2i p;
      p.x = (r0.x0 + r0.x1)/2;
      p.y = r0.y1;

      Graph_node* new_node = new Graph_node();
      new_node->p.x = p.x;
      new_node->p.y = p.y;

      new_node->n1 = n;
      new_node->n2 = NULL;

      pts.push_back(new_node);
	  
      new_node->index = all_nodes.size();
      all_nodes.push_back(new_node);
    }
#endif
    FORALL(n->nl[2],it2) {
      Decomp::Node *neighbour = *it2;

      if (neighbour->type!=Decomp::Node::R_NODE) {
	cout<<"??"<<endl;
	continue;
      }

      if (neighbour->data != n->data || n->data ==1 || neighbour->data ==1) {
	continue;
      }

      //handle it differently if the neighbour rectangle is visited
      if (neighbour->visited) {
	Graph_node *already_exist = get_node(n,neighbour);
	if (already_exist==NULL) {
	  ERR("should not be happening");
	}

	pts.push_back(already_exist);
	continue;
      }

      Decomp::Rect r1 = neighbour->r;

      Point2i p1,p2,p3,p4;

      p1.x = r0.x0;
      p1.y = r0.y0;
      p2.x = r0.x1;
      p2.y = r0.y0;

      p3.x = r1.x0;
      p3.y = r1.y1;
      p4.x = r1.x1;
      p4.y = r1.y1;
		    
      Point2i p = get_point(p1,p2,p3,p4);

      Graph_node* new_node = new Graph_node;

      new_node->p = p;
      new_node->n1 = n;
      new_node->n2 = neighbour;

      pts.push_back(new_node);

      //add it to the list if it is not added yet
      new_node->index = all_nodes.size();
      all_nodes.push_back(new_node);
    }

    //west neighbour
#if 0
    if (n->nl[3].empty()) {
      Point2i p;
      p.x = r0.x0;
      p.y = (r0.y0 + r0.y1)/2;

      Graph_node* new_node = new Graph_node();
      new_node->p.x = p.x;
      new_node->p.y = p.y;

      new_node->n1 = n;
      new_node->n2 = NULL;

      pts.push_back(new_node);
	  
      new_node->index = all_nodes.size();
      all_nodes.push_back(new_node);
    }
#endif
    FORALL(n->nl[3],it3) {
      Decomp::Node *neighbour = *it3;

      if (neighbour->type!=Decomp::Node::R_NODE) {
	cout<<"??"<<endl;
	continue;
      }

      if (neighbour->data != n->data || n->data ==1 || neighbour->data ==1) {
	continue;
      }

      //handle it differently if the neighbour rectangle is visited
      if (neighbour->visited) {
	Graph_node *already_exist = get_node(n,neighbour);
	if (already_exist==NULL) {
	  ERR("should not be happening");
	}

	pts.push_back(already_exist);
	continue;
      }
		
      Decomp::Rect r1 = neighbour->r;

      Point2i p1,p2,p3,p4;

      p1.x = r0.x0;
      p1.y = r0.y0;
      p2.x = r0.x0;
      p2.y = r0.y1;

      p3.x = r1.x1;
      p3.y = r1.y1;
      p4.x = r1.x1;
      p4.y = r1.y0;
		    
      Point2i p = get_point(p1,p2,p3,p4);

      Graph_node* new_node = new Graph_node;

      new_node->p = p;
      new_node->n1 = n;
      new_node->n2 = neighbour;

      pts.push_back(new_node);

      //add it to the list if it is not added yet
      new_node->index = all_nodes.size();
      all_nodes.push_back(new_node);
    }

#if 0
    //interconnect all nodes
    FORALL(pts,i) {
      Graph_node *node = *i;
		
      FORALL(pts,j) {
	Graph_node *another = *j;
	if (node!=another) {

	  if (!node->has_neighbour(another)) {
	    node->neighbours.push_back(another);
	    another->neighbours.push_back(node);
			
	    //cout<<"connecting:("<<node->p.x<<","<<node->p.y<<")-->("<<another->p.x<<","<<another->p.y<<")\n";
	  }
	}
      }
    }
#endif

    FORALL(pts,i) {
      if (i!=pts.begin()) {

	Graph_node *node = *i;
		
	node->neighbours.push_back(pts.front());

	pts.front()->neighbours.push_back(node);
      }
    }

    pts.clear();
    
    while(!pts.empty()) {
      pts.pop_back();
    }

    //mark the current as visited
    n->visited