#ifndef Game1_ST_ForceFieldMove_H
#define Game1_ST_ForceFieldMove_H

#include "Game1_SimpleTerrain.H"

namespace Game1_SimpleTerrain {

  /** Two Dimensional Vector */
  class Vec_2D{
  public:
    sint4 x;/// x-coord
    sint4 y;/// y-coord
    Vec_2D(sint4 x_=0, sint4 y_=0) :  x(x_), y(y_){};
    Vec_2D(const Vec_2D &other) :  x(other.x), y(other.y){};
    Vec_2D& operator=(const Vec_2D &other) { 
      if( this != &other ) {
        x = other.x;
        y = other.y;
      }
      return *this;
    };
    Vec_2D(const Loc &a, const Loc &b) : x(b.x - a.x), y(b.y - a.y) {};
    Vec_2D operator+(const Vec_2D &other) const; ///Vector addition
    Vec_2D operator-(const Vec_2D &other) const; ///Vector subtraction
    Vec_2D project_onto(const Vec_2D &other) const; /// Project the vector onto the line defined by the given vector
    sint4 dot_prod(const Vec_2D &other) const;
    real8 length() const;
    Vec_2D multiply(real8 k) const;
    bool operator==(const Vec_2D &other) const;
    bool operator!= (const Vec_2D &other) const;

    inline static sint4 round(real8 real) {
      return (sint4)floor(real + 0.5);
    }
  };

  /// Simple Movement-execution module
  class ST_ForceFieldMove {
  public:
    ST_ForceFieldMove(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 client_pID_, sint4 neutral_pID_);
    ~ST_ForceFieldMove();

    /// add command-line options specific to this implementation, using static methods of class Options
    static void add_options(void);

    /// Monitor task execution, return move commands and status messages.
    void execute_tasks(ST_Terrain::Tasks &tasks, 
                       ST_Terrain::Objects &objsToStop,
                       Vector<MoveCmd> &cmds, 
                       Vector<StatusMsg> &msgs);

    //---------------------------------------------------------------------
    /// \name Changes to the world.
    //---------------------------------------------------------------------
    //@{
    void insert_object(const Object *obj);
    void remove_object(const Object *obj);
    void update_object(const Object *obj);
    void insert_boundary(const Segment& line);
    //@}
    //---------------------------------------------------------------------

    // public to allow REGISTER_TYPEOF definitions
    typedef std::map<const Object*, Loc>  Obj2LocMap;
    typedef std::map<Loc, bool> SegmentEndsMap;

  private:

    //---------------------------------------------------------------------
    /// \name Internal accessories
    //---------------------------------------------------------------------
    //@{
    /** Test if the object is at the given location, 
	accounting for a 1-tick look-ahead for moving objects. */
    static bool is_at_location(const Object* obj, const Loc& target);
    /** Test if the distance from the object to the target is within the given range.*/
    static bool is_near_location(const Object* obj, const Loc& target, sint4 range);
    /** Test if the object is at the goal of the given task, implementing 
        (some) of the various definitions of 'goal'. */
    bool is_at_goal(const Object* obj, const Task& task, const ST_Terrain::TIter &st_task);

#if 0
    /** Test if the object is within firing range */
    static bool is_in_range(const Object* moving_obj, const Object *goal_obj);
    /** Test if the location is within firing range */
    static bool is_in_range(const Object* moving_obj, const Loc goal_loc);
#endif 

    /** Add a move command to the Vector of commands, telling the object to 
	move to the target location at maximum speed.
	Also, updates the location associated with the object for forcefield purposes
    */
    void add_move_command(const Object *obj, Loc target, Vector<TerrainBase::MoveCmd> &cmds);
    /** Determines what team an object is on.
	If the team cannot be determined, returns -1 */
    sint4 what_team(const Object *obj);
    //@}
    //---------------------------------------------------------------------


    //---------------------------------------------------------------------
    /// \name Force Field methods
    //--