// $ Id $

#include "Game1_accessories.H"
#include <algorithm>

#define TEST(x)
//#define TEST(x) x

using namespace std;

/*********************************************************************************/
/** positive if r is on the left of the directed segment from p to q;
    negative if r is on its right;
    0 if r is on the line containing the segment. */
static Scalar is_left_turn(const ScalarPoint &p, 
                           const ScalarPoint &q, 
                           const ScalarPoint &r) 
{ 
  // adapted for the reversed y-axis used by ORTS
  return (q.y-p.y)*(r.x-p.x)-(q.x-p.x)*(r.y-p.y);
}

/*********************************************************************************/
void compute_roundabout(const Vector<ScalarPoint> &points, 
                        const ScalarPoint &reference,
                        Vector<ScalarPoint> &result)
{
  result.clear();
  sint4 n = points.size();

  TEST( cout << "reference = [" << reference.x << "," << reference.y << "];" << endl;
        cout << "points = [" << endl;
        FORALL(points, i) cout << i->x << "," << i->y << ";" << endl;
        cout << "];" << endl; );

  // degenrate case: S contains 0/1 points, return result=S
  if( n <= 1 ) {
    result = points;
    return;
  }   

  // (1) Find the left most and right most points in S, relative to P

  // start with 2 arbitrary points, check which is on the left of the other
  ScalarPoint leftmost(points[0]);
  ScalarPoint rightmost(points[1]);
  if( is_left_turn(reference, leftmost, rightmost) > 0 ) // left turn is wrong
    swap(leftmost, rightmost);
  // incrementally update after considering all other points
  for( int i = 2; i < n ; i++ ) {
    // is on the left of leftmost?
    if( is_left_turn(reference, leftmost, points[i]) > 0 )
      leftmost = points[i];
    // else, is on the right of rightmost?
    else if( is_left_turn(reference, rightmost, points[i]) < 0 )
      rightmost = points[i];
  }

  TEST( cout << "leftmost = [" << leftmost.x << "," << leftmost.y << "];" << endl;
        cout << "rightmost = [" << rightmost.x << "," << rightmost.y << "];" << endl; );

  // (2) To get each consecutive point on the convex hull, repeatedly find
  //     the leftmost point relative to the previous point. 
  //     This is theoretically slow - O(n^2) - but should do for the small sets
  //     (n=10) we are working with.
  do {
    ScalarPoint current(leftmost);
    result.push_back(current);
    // test all other points as candidates for being the 
    // leftmost relative to the current points
    int start = 0;
    if( current == points[0] ) start = 1;
    leftmost = points[start];
    for( int i = start+1; i < n ; i++ ) {
      // is on the left of leftmost?
      if( points[i] != current && is_left_turn(current, leftmost, points[i]) > 0 )
        leftmost = points[i];
    }
  } while( leftmost != rightmost );
  result.push_back(rightmost);

  TEST(
    cout << "roundabout = [" << endl;
    FORALL(result, i) cout << i->x << "," << i->y << ";" << endl;
    cout << "];" << endl;
  );
}

/*********************************************************************************/
static void bisector_direction(const ScalarPoint &p, 
                               const ScalarPoint &q, 
                               const ScalarPoint &r,
                               real8 &a, real8 &b)
{
  // The normals to the lines == (a,b) in the line equation ax+by+c=0
  ScalarPoint pq_normal(q.y-p.y, -(q.x-p.x));
  ScalarPoint qr_normal(r.y-q.y, -(r.x-q.x));
  // Normalize the normals to unit length
  real8 n1 = sqrt((real8)pq_normal.x*pq_normal.x + pq_normal.y*pq_normal.y);
  real8 n2 = sqrt((real8)qr_normal.x*qr_normal.x + qr_normal.y*qr_normal.y);
  // add the normals to get the bisector direction
  a = n2 * pq_normal.x + n1 * qr_normal.x;
  b = n2 * pq_normal.y + n1 * qr_normal.y;

  // If this produces a degenerate line.
  if (a == 0 && b == 0) {
    a = -(n2 * pq_normal.y - n1 * qr_normal.y);
    b = n2 * pq_normal.x - n1 * qr_normal.x;
  }
}

/*********************************************************************************/
void blowup_polygon(const Vector<ScalarPoint> &polygon, real8 distance, Vector<ScalarPoint> &result)
{
  // For each two consecutive segments, compute the bisector and
  // move the joint vertex outward in this direction
  int n = polygon.size();
  assert( n >= 3 && "A polygon must have at least 3 vertices");
  result = polygon;
  for(int i=0; i < n; i++) {
    real8 a,b;
    // first point
    if( i == 0 )
      bisector_direction(polygon[n-1], polygon[0], polygon[1], a, b);
    // last point
    else if( i == n-1 )
      bisector_direction(polygon[n-2], polygon[n-1], polygon[0], a, b);
    // all other points
    else
      bisector_direction(polygon[i-1], polygon[i], polygon[i+1], a, b);

    // normalize direction
    real8 inv_norm = 1.0/sqrt(a*a + b*b);
    a *= inv_norm;
    b *= inv_norm;

    // move vertex
    result[i].add( ScalarPoint(FixedPoint::from_real8(distance*a), 
                               FixedPoint::from_real8(distance*b)) );
  }
}
