
/** @file Worlded.C
    World representation for map editor.
*/

// This is an ORTS file
// licensed under the GPL

#include "Worlded.H"
//#include "../../../libs/serverclient/src/RandomIsland.C"
Worlded* Worlded::worldedInstance = NULL;
bool     Worlded::initialized = false;
using namespace std;

struct Point{
  sint4 x;
  sint4 y;
};

void Worlded::add_options()
{
  Options o("World Editor");
  
  o.put("-loadmap", std::string(""),      "map file name to load");
  o.put("-newt",    true,            "new terrain");
  o.put("-debug",   false,           "debug mode");
 
// FRANO
  o.put("-frano",false,"frano's test map");

  o.put("-disp",    "2d display");

  o.put("-m",       std::string("server"),"mode: server|client|game|test|worker");
  o.put("-showv",   false,           "show views");
  o.put("-showl",   false,           "show message length");

  o.put("-hfile",   std::string(""),      "create map from height/ramp file");
  o.put("-out",     std::string(""),      "save map to file");

}

Worlded* Worlded::instance(){
  if(worldedInstance == NULL)
    worldedInstance = new Worlded();

  return worldedInstance;
}


void Worlded::initialize(sint4 t_h, sint4 rad, Tile::Type t)
{
  if(!initialized){
  map_data = new Map<Tile>;
  units = new Vector<Unit>;
  radius = rad;
  tile_height = t_h;

  add_options();
  MapTool::Options::add();

  sint4 x, size = 400;
  bool water = true;
  Options::get("-x", x);
  sint4 y;
  Options::get("-y", y);
  Options::set("-island", size);
  Options::set("-water", water);
  sint4 seed;
  Options::get("-seed", seed);
  rand.srand(seed);
  map_data->allocate(x, y);
  tile_type = t;
  initialized = true;
  }
}

bool Worlded::isInitialized() {
  return initialized;
}


//change tiles around x, y to tile type and increment height by diff
void Worlded::change_tiles(sint4 x, sint4 y){
  sint4 width;
  Options::get("-x", width);
  sint4 height;
  Options::get("-y", height); 
  if((x - radius) < 0 || (y < radius) < 0 || (x + radius) >= width || (y + radius) >= height)
    return;
  x-=radius;
  y-=radius;
  radius+=radius;
  FORS(i, radius){
    FORS(j, radius){
      (*map_data)(x + i, y + j).set_type(tile_type);
      (*map_data)(x + i, y + j).set_flat(tile_height);
    }
  }
}

//Places a unit with given name in the world, no safety checking on x and y

void Worlded::place_unit(std::string n, sint4 x, sint4 y) {
  Unit u;
  u.name = n;
  u.x = x;
  u.y = y;
  units->push_back(u);
}

void Worlded::set_tile(Tile::Type t) { tile_type = t; }
Tile::Type Worlded::get_type() { return tile_type; }

void Worlded::set_map(Map<Tile>* m) { map_data = m;}
Map<Tile>* Worlded::get_map() { return map_data; }

void Worlded::set_height(sint4 h) { tile_height = h; }
sint4 Worlded::get_height() { return tile_height; }

void Worlded::set_radius(sint4 r) { radius = r; }
sint4 Worlded::get_radius() { return radius; }

void Worlded::set_units(Vector<Unit>* u) { units = u; }
Vector<Worlded::Unit>* Worlded::get_units() { return units; }
  
void Worlded::generate_map()
{
  // note: if loading a map, -nplayers values must match
  sint4 num_players;
  
  Options::get("-nplayers", num_players);
 
  sint4 size;
  Options::get("-island", size);
  sint4 tiles_x, tiles_y; 
  Options::get("-x", tiles_x);
  Options::get("-y", tiles_y);
  sint4 tile_n = tiles_x * tiles_y;

  
  map_data->allocate(tiles_x, tiles_y);
  sint2 *o_safe = new sint2[tile_n]; // valid square if 0
  if (!o_safe) ERR("out of memory");
  memset(o_safe, 0, tile_n*sizeof(*o_safe));
  real8 convex = 0;
  Options::get("-convex", convex); 
  sint4 r;
  r = rand.rand_uint4();
  //radius of inner disk
  sint4 rad_i = tiles_x /8;
  //positions of players
  sint4 *x_pos = new sint4[num_players];
  sint4 *y_pos = new sint4[num_players];
  bool *flags = new bool[tile_n];
  bool *checked = new bool[tile_n];
  if (!x_pos || !y_pos || !flags || !checked) ERR("out of memory");
  memset(x_pos, 0, num_players*sizeof(*x_pos));
  memset(y_pos, 0, num_players*sizeof(*y_pos));
  memset(flags, 0, tile_n*sizeof(*flags));
  memset(checked, 0, tile_n*sizeof(*checked));

  Vector<Po