// $Id: GfxObject.C,v 1.43 2007/03/21 17:16:02 orts_mburo Exp $

#include "GfxObject.H"
#include "GfxModule.H"
#include "GameStateModule.H"
#include "GfxTerrain.H"
#include "GfxTerrainTile.H"
#include "GameTile.H"
#include "Game.H"
#include "PlayerInfo.H"
#include "GameObj.H"
#include "Minimap.H"
#include "GfxEmitter.H"

using namespace std;

const sint4 AIR_OBJ_HEIGHT = 3; // in tiles (was 4)

uint4 GfxObject::id0 = 2; // > 1
GfxObject::Id2GfxObjMap GfxObject::id2obj_map;
GfxTerrain* GfxObject::terrain = 0;
GfxModule* GfxObject::gfxm = 0;

GfxGameObject::Gfx2GfxGameObj GfxGameObject::gfx2gfxgame_map;
GfxGameObject::Gfx2GameObj    GfxGameObject::gfx2game_map;

//===================================================================

GfxObject::GfxObject()
{
  drawn = 0;
  my_id = ++id0;
  if (my_id >= (2000000000 >> SUB_ID_LOG)) ERR("id out of bounds"); // fixme: recycle ids?
  id2obj_map[my_id] = this;
  selected = false;
  dead = false;
  moved = true;
  zcat = O_ON_LAND;

  poses.resize(1);
}

//-------------------------------------------------------------------

GfxObject::~GfxObject()
{
  FIND (id2obj_map, i, my_id);
  if (i != id2obj_map.end()) id2obj_map.erase(i);
  FORALL (sub_objects, it) delete *it;
  FORALL (emitters, it) delete *it;
}

//===================================================================

void GfxObject::match_id(uint4 id)
{
  FIND (id2obj_map, i, my_id);
  if (i != id2obj_map.end()) id2obj_map.erase(i);
  my_id = id;
}

//===================================================================

void GfxObject::draw_at(const Mat4<real4> &mat)
{
  glPushMatrix();
  glMultMatrixf(mat.ptr());
  draw();
  glPopMatrix();
}

void GfxObject::draw_emitters_at(const Mat4<real4> &mat)
{ 
  glerror("GfxObject::draw_emitters_at 0");
  
  glPushMatrix();  
  glMultMatrixf(mat.ptr());
  
  draw_emitters();

  glPopMatrix();
  glerror("GfxObject::draw_emitters_at 1");
}

//===================================================================

void GfxObject::draw_emitters()
{
  glerror("GfxObject::draw_emitters 0");
  
  glPushMatrix();
  glTranslatef(pose.center.x, pose.center.y, pose.center.z);
  glRotatef(pose.y_rot, 0, 1, 0);

  FORALL (sub_objects, i) (*i)->draw_emitters();
  FORALL (emitters, i)    (*i)->draw_emitters();

  glPopMatrix();
  glerror("GfxObject::draw_emitters 1");
}

//===================================================================

void GfxObject::attachGfxObject(GfxObject *o)
{
  sub_objects.push_back(o);
}

//-------------------------------------------------------------------

void GfxObject::attachEmitter(GfxEmitter *o)
{
  emitters.push_back(o);
}

//===================================================================

void GfxObject::update(real4 dt)
{
  /*
  aabb_o.x = center.x - radius;
  aabb_o.y = center.y - radius;
  aabb_o.z = center.z - radius;
  aabb_d.x = aabb_d.y = aabb_d.z = 2*radius;
  */

  FORALL (sub_objects, it) (*it)->update(dt);
  FORALL (emitters, it)    (*it)->update(dt);
}

//===================================================================

void GfxObject::set_position(R x, R y, R z, uint4 p)
{
  poses[p].center.x = x;
  poses[p].center.y = y;
  poses[p].center.z = z;

  pose.center = Vec3<R>(0,0,0);
  FORALL (poses, i) pose.center += (*i).center;
  moved = true;
}

//-------------------------------------------------------------------

void GfxObject::set_rotation(real4 y, uint4 p)
{
  poses[p].y_rot = y;

  pose.y_rot = 0;
  FORALL (poses, i) pose.y_rot += (*i).y_rot;
  moved = true;
}

//-------------------------------------------------------------------

void GfxObject::set_pitch(real4 x, uint4 p)
{
  poses[p].x_rot = x;

  pose.x_rot = 0;
  FORALL (poses, i) pose.x_rot += (*i).x_rot;
  moved = true;
}

//-------------------------------------------------------------------

void GfxObject::set_scale(real4 s, uint4 p)
{
  poses[p].scale = s;
  pose.scale = 1.0;
  FORALL (poses, i) pose.scale *= (*i).scale;
  moved = true;
}

//===================================================================

void GfxObject::apply_pose()
{
  glTranslatef(pose.center.x, pose.center.y, pose.center.z);
  glRotatef(pose.y_rot, 0, 1, 0);
  glRotatef(pose.x_rot, 0, 0, 1);
  glScalef(pose.scale, pose.scale, pose.scale);
}

//===================================================================

GfxObject* GfxObject::id2obj(uint4 id)
{
  FIND (id2obj_map, it, id);
  if (it != id2obj_map.end()) return it->second;
  return 0;
}

//===================================================================
//===================================================================
//===================================================================

GfxGameObject::GfxGameObject(GameObj *g) : gob(g)
{
  gfx_obj = 0;
  pose_num = -1;
  y_rot = 0.0;
  x_rot = 0.0;
  heading[0] = 0.0;
  pitch[0] = 0.0;
}

//-------------------------------------------------------------------

GfxGameObject::~GfxGameObject()
{
  FIND (gfx2game_map, it, gfx_obj);
  if (it != gfx2game_map.end()) gfx2game_map.erase(it);

  FIND (gfx2gfxgame_map, it2, gfx_obj);
  if (it2 != gfx2gfxgame_map.end()) gfx2gfxgame_map.erase(it2);
}

//===================================================================

void GfxGameObject::attach_gfxobj(GfxObject *gfx)
{
  pose_num = gfx->poses.size();
  gfx->poses.resize(pose_num+1);

  gfx_obj = gfx;
  gfx2gfxgame_map[gfx] = this;
  gfx2game_map[gfx] = gob;

  game_tick();
}

//===================================================================

void GfxGameObject::game_tick()
{
  assert(this->gob);
  GameObj &gob = *this->gob;

  const sint4 TP  = GUI_gfxm->get_game().get_tile_points();
  const real4 ITP = 1.0/TP;
  
  sint4 ox = *gob.sod.x;
  sint4 oy = *gob.sod.y;
  sint4 x = ox;
  sint4 y = oy;

  if (gfx_obj) {
    if (x < 0) {
      if (gfx_obj->hidden < 2) gfx_obj->hidden = 3;
      return;
    } else if (gfx_obj->hidden) {
      gfx_obj->hidden = 1;
      interp = 0; // don't slide to the new position
    }
  }

  const Map<GameTile> &map = GUI_gfxm->get_game().get_map();
  sint4 mw = map.get_width();
  sint4 mh = map.get_height();  
  
  if (x < 0) x = 0;
  else if (x >= mw * TP) x = mw * TP - 1;

  if (y < 0) y = 0;
  else if (y >= mh * TP) y = mh * TP - 1;

  const GameTile &tile = map(x / TP, y / TP);
  real4 inv_h = 1.0 / GameConst::HEIGHT_MULT;
  real4 h = inv_h * tile.get_min_h();
  real4 h_diff;
  real4 noise = 0.f;
 
  position[1] = position[0];
  heading[1]  = heading[0];
  pitch[1]    = pitch[0];
  
  switch (*gob.sod.zcat) {

  case Object::UNDER_WATER:
    if (gfx_obj) gfx_obj->zcat = GfxObject::O_ON_LAND;
    --h;
    break;

  case Object::ON_WATER:
    if (gfx_obj) gfx_obj->zcat = GfxObject::O_ON_LAND;
    break;

  case Object::ON_LAND:
  
    if (gfx_obj) {
      const double M_180_PI = 1/0.01745329251994329576;	//!< \f$ 180 / \pi \f$
      //!< The smallest X of type float such that 1.0 + X != 1.0.
      const double M_EPSILON = 0.00001;
    
      gfx_obj->zcat = GfxObject::O_ON_LAND;
      
      enum TILT { TILT_NONE=0, 
                  TILT_RAMPS=1, 
                  TILT_NOISE=2, 
                  TILT_MODE_SINGLE=4,
                  TILT_MODE_LINEAR=16 
                };
      sint4 tilt = 0;
      sintptr *c = gob.get_int_ptr("tilt_mode");
      if (c) tilt = *c;
      
      if(tilt & TILT_NONE)
      {
        pitch[0] = 0.0f;
        
        // Height of the terrain directly below us     
        h = height_above(&noise);
        h += noise;
        break;
      }
      else if(tilt & (TILT_RAMPS | TILT_NOISE))
      {
        // Get the tile sample range
        c = gob.get_int_ptr("tilt_range");
        real4 range = 2.f;
        if (c) range = static_cast<real4>(*c);
        
        // Sample 3 colinear points
        if(tilt & TILT_MODE_LINEAR)
        {          
          // Height of the terrain in front and behind the object
          real4 noise_front, noise_back;
          real4 h_front = height_above(&noise_front, range);
          real4 h_back  = height_above(&noise_back, -range);   

          // Set the objects height to the center position
          h = height_above(&noise); 
          h += noise;

          if(!(tilt & TILT_RAMPS))
            h_front = h_back;

          if(tilt & TILT_NOISE)
          {
            h_front += noise_front;
            h_back  += noise_back;
          }

          // Set the object at a weighted combination of the sample heights
          h = (2.f*h + h_front + h_back) / 4.f;
          
          h_diff = (h_front - h_back) * TP;
        
        // Sample a point ahead of the center (TILT_MODE_SINGLE)
        }else
        {
          // Height of the terrain directly below us     
          h = height_above(&noise);

          // Get the tile sample range
          c = gob.get_int_ptr("tilt_range");
          real4 range = 2.f;
          if (c) range = static_cast<real4>(*c);

          // Height of the terrain in front of the object
          real4 h_ahead, noise_ahead;
          h_ahead = height_above(&noise_ahead, range);

          if(!(tilt & TILT_RAMPS))
            h_ahead = h;

          if(tilt & TILT_NOISE)
            h_ahead += noise_ahead;
          else
            h_ahead += noise;

          h += noise;    // Adjust for noise in the terrain

          h_diff = (h_ahead - h) * TP;
        }

        if (abs(h_diff) > M_EPSILON)
          // Pitch to the slope of the terrain under us
          pitch[0] = atan2(h_diff, range) * M_180_PI;
        else
          pitch[0] = 0.0f;
      }
    }
    break;

  case Object::IN_AIR:
    if (gfx_obj) gfx_obj->zcat = GfxObject::O_IN_AIR;
    h = AIR_OBJ_HEIGHT;
    break;

  default: ERR("unknown zcat"); break;
  }
  
  //------------------------------

  Object::Shape shape = static_cast<Object::Shape>(*gob.sod.shape);

  if (shape == Object::CIRCLE) {

    position[0].x = ox * ITP;
    position[0].y = h;
    position[0].z = oy * ITP;

    //    real4 radius = *gob.sod.radius * ITP;
    //    aabb_o.x = position[0].x - radius;
    //    aabb_o.y = position[0].y - radius;
    //    aabb_o.z = position[0].z - radius;
    //    aabb_d.x = aabb_d.y = aabb_d.z = 2*radius;

    sintptr *face_target = gob.get_int_ptr("face_target");
    
    heading[0] = 0;
    sintptr *ld = gob.get_int_ptr("last_direction");  // 0 = not attacking, 1...GameConst::ATTACK_N
    if ((!face_target || *face_target) &&
        ld && (*ld || gob.attr_changed("last_direction"))) {
      // if the unit has stopped shooting, keep looking in the same direction
      sint4 foo = *ld ? *ld : gob.get_int_prev("last_direction");
      heading[0] = GameConst::angle_from_dir(foo, GameConst::ATTACK_N);
    } else {
      const sintptr *hd = gob.get_int_ptr("heading");  // 0 = stopped, 1...GAME_CONST::HEADING_N
      if (hd) heading[0] = GameConst::angle_from_dir(*hd, GameConst::HEADING_N);
    }

  } else if (shape == Object::RECTANGLE) {

    heading[0] = 0;

    //    sint4 width  = gob.get_int("width");
    //    sint4 height = gob.get_int("height");

    //    aabb_o.x = (ox-width*0.5) * ITP;
    //    aabb_o.y = h;
    //    aabb_o.z = (oy-height*0.5) * ITP;

    //    aabb_d.x = width  * ITP;
    //    aabb_d.z = height * ITP;
    //    aabb_d.y = (aabb_d.z + aabb_d.x) * 0.5;

    position[0].x = ox * ITP;
    position[0].y = h;
    position[0].z = oy * ITP;

  } else {

    ERR2("unknown shape ", shape);

  }

  if (interp > 0) interp = 0;
  else {
    //first time
    interp      = 0;
    position[1] = position[0];
    heading[1]  = heading[0];
    pitch[1]    = pitch[0];
  }
}

real4 GfxGameObject::height_above(real4 *noise, real4 x_offset)
{
  const double M_PI_180 = 0.01745329251994329576;	  //!< \f$ \pi /180 \f$
  const sint4 TP  = GUI_gfxm->get_game().get_tile_points();
  const real4 ITP = 1.0/TP;
  
  const Map<GameTile> &map = GUI_gfxm->get_game().get_map();
  
  GameObj &gob = *this->gob;  
  sint4 x = *gob.sod.x;
  sint4 y = *gob.sod.y;
  GfxTerrainTile *gtt = GUI_gfxm->terrain->tile(x / TP, y / TP);
  //  sint1 *f = gtt->flattened;

  sint4 mw = map.get_width();
  sint4 mh = map.get_height();  
  
  if (x < 0) x = 0;
  else if (x >= mw * TP) x = mw * TP - 1;

  if (y < 0) y = 0;
  else if (y >= mh * TP) y = mh * TP - 1;

  // todo: apply y_offset the test location 

  sint4 sx = x + static_cast<sint4>((x_of