// $Id: Gfx2d.C,v 1.2 2006/08/09 03:14:07 nathanie Exp $
 
// This is an ORTS file 
// (c) Michael Buro  
// (c) Tim Furtak    
// (c) Keith Yerex   
// licensed under the GPL

// 2d stuff and script gfx interface

#include "GfxGlobal.H"
#include "Global.H"
#include "GameStateModule.H"
#include "GameObj.H"
#include "GfxModule.H"
#include "Game.H"
#include "PlayerInfo.H"
#include "GfxObject.H"
#include "GfxTerrain.H"
#include "GfxAnimesh.H"
#include "Minimap.H"
#include "GameTile.H"
#include "Token.H"
#include "LayoutHV.H"
#include "Options.H"
#include "Image.H"
#include "GfxModule.H"
#include "ScriptModel.H"
#include "ScriptVector.H"

using namespace std;


#define PARSE_ERR(s) {\
  errstr << endl << ": parse error at line " << tokenizer.cur_line();\
  ERR(s);\
}

//===================================================================

stype GfxModule::GUI_CreateView(const SActParam &)
{
  sint4 id;
  for (id = 1; id <= sint4(views.size()) + 1; ++id) {
    bool found = true;
    FORALL (views, i) {
      if ((*i).id == id) {
        found = false;
        break;
      }
    }
    if (found) break;
  }

  SDL_Rect r;
  r.x = 0;
  r.y = 0;
  r.w = 100;
  r.h = 100;

  GfxView v(r);
  v.id = id;
  
  if (get_game().is_rotated()) { v.x_angle = 65; v.y_angle = 45; }
  else                         { v.x_angle = 65; v.y_angle = 0;  }
  
  //  v.frame_color = view_colors[views.size()+1];
  v.frame_color = view_colors[1];
  
  v.vp_scale = max(v.viewport.w, v.viewport.h);
  
  cout << "NEW VIEW: "; v.write(); 
  views.push_back(v);

  update_active_view();
  return stype(id);
}

//-------------------------------------------------------------------

stype GfxModule::GUI_DeleteView(const SActParam &p)
{
  VALN_CHECK(1);  
  sint4 id = p.val[0].to_int();

  cout << "DELETING VIEW: " << id << endl;

  FORALL (views, i) {
    if ((*i).id == id) {
      *i = views[views.size()-1];
      views.resize(views.size()-1);
      update_active_view();
      return stype();
    }
  }
  return stype(1);
}

//-------------------------------------------------------------------

stype GfxModule::GUI_ResizeView(const SActParam &p)
{
  VALN_CHECK(5);  
  
  sint4 k = p.val[0].to_int();
  sint4 id = -1;

  FORALL (views, i) {
    if ((*i).id == k) {
      id = i-views.begin();
      break;
    }
  }

  if (id == -1) return stype(1);

  GfxView &v = views[id];

  v.viewport.x = p.val[1].to_int();
  v.viewport.y = p.val[2].to_int();
  v.viewport.w = p.val[3].to_int();
  v.viewport.h = p.val[4].to_int();

  v.vp_scale = max(v.viewport.w, v.viewport.h);
  update_active_view();
  return stype();
}

//===================================================================

void render_bitmap_string(GLfloat x, GLfloat y, void *font, const char *text)
{
  const sint4 lineHeight = 13;
    
  const char *c;
  glRasterPos2f(x,y);
  for (c=text; *c != '\0'; ++c) {
    if (*c == '\n') {
      y -= lineHeight;
      glRasterPos2f(x,y);
      continue;
    }
    glutBitmapCharacter(font, *c);
  }
}

//===================================================================

/** Negate MODELVIEW rotation and scaling so that objects are drawn facing the camera.
    Used for things like health bars.
*/

void GfxModule::overlay_mode()
{
  const GfxView &v = *cur_view;

  real4 modelview[16];
  glGetFloatv(GL_PROJECTION_MATRIX, modelview);

  real4 s1, s2, s3;

  if (opt.perspective_mode) {
    s1 = 8.0;
    s2 = 8.0;
    s3 = 8.0;
  } else {
    real8 m = min(modelview[0], modelview[5]);
    const real8 cap = 0.08;

    if (m > cap) m = cap;
    m = cap / m;
    s1 = s2 = s3 = 8.0 * m;
  }

  glGetFloatv(GL_MODELVIEW_MATRIX, modelview);

  // remove rotation, negate projection matrix scaling by adjusting modelview
  modelview[0] = s1;   modelview[4] = 0.0;  modelview[8]  = 0.0;
  modelview[1] = 0.0;  modelview[5] = s2;   modelview[9]  = 0.0;
  modelview[2] = 0.0;  modelview[6] = 0.0;  modelview[10] = s3;

  glLoadMatrixf(modelview);

  real4 s =  1.0 / v.vp_scale * 3;
  
  glScalef(s,s,1.0);
  
#if 0
  if (!perspective_mode) {
    sint4 d = v.viewport.w - v.viewport.h;
    if     