// $Id: OSL_Parser.C,v 1.3 2004/11/21 22:56:10 orts_mburo Exp $

// (c) Jonathan Yip, licensed under the GPL

#include "OSL_Parser.H"
#include "OSL_RecStream.H"
#include "Lexer.H"

using namespace std;

#define DEBUG 1

#define DEBUGSTR errstr << "DEBUG function: " 

#define PARSE_ERR(s) {                                                  \
  errstr << token_stream[token_stream_ptr].file << "(" << token_stream[token_stream_ptr].line_num << "): " << s \
         << endl; exit(20);                                             \
}

#define PARSE_ERR2(s1, s2) {                                                  \
  errstr << token_stream[token_stream_ptr].file << "(" << token_stream[token_stream_ptr].line_num << "): " << s1 << " " << s2 \
         << endl; exit(20);                                             \
}

/*
#define PARSE_ERR2(s1, s2) {                                            \
  errstr << lex->cur_file() << "(" << lex->cur_line_num() << "): "      \
         << s1 << " " << s2 << endl; exit(20);                          \
}
*/

#define PARSE_ERR4(file, line, pos, s) {                                \
  errstr << file << "(" << line << ":" << pos << "): " << s             \
	 << endl; exit(20);                                             \
}

#define PARSE_ERR5(file, line, pos, s1, s2) {                           \
  errstr << file << "(" << line << ":" << pos << "): " << s1 << " "     \
	 << s2 << endl; exit(20);                                       \
}


#define PROGRAM_ERR ERR("Should not reach here, error in C++ program")

namespace OSL {

  Parser::Parser(Program &p, const string &file_name) : prog(p) {
    lex = new Lexer(rs);
    FILE *fp = fopen(file_name.c_str(), "r");
    
    if (!fp) ERR2("File not present:", file_name);
    rs.push(fp, file_name);
    
    curr_class = 0;
    curr_func = 0;
    curr_static = false;
  }

  inline const Token Parser::get_token() {
    Token tk;

    try {
      tk = lex->get_token();
      // fixme pos number
      token_stream.push_back( TokenInfo(tk, lex->cur_line_num(),
					0, lex->cur_file()) );
      tk.write(); cout << endl;
    }
    catch (Lexer::SyntaxErrorException &e) {
      cout << "Syntax error: " << e.msg << endl;
    }
    catch (RecStream::ReadErrorException &e) {
      cout << "Read error: " << e.msg << endl;
    }
    return tk;
  }
  
  inline const Token Parser::peek_token() {
    Token tk;

    try {
      tk = lex->peek_token();
    }
    catch (Lexer::SyntaxErrorException &e) {
      cout << "Syntax error: " << e.msg << endl;
    }
    catch (RecStream::ReadErrorException &e) {
      cout << "Read error: " << e.msg << endl;
    }
    return tk;
  }

  // fixme
  inline const Token Parser::get_token_info() {
    token_stream[token_stream_ptr].tk.write();
    cout << endl;
    return token_stream[token_stream_ptr++].tk;
  }

  inline const Token Parser::peek_token_info() {
    return token_stream[token_stream_ptr].tk;
  }
  
  inline Parser::var_location Parser::find_variable(const string var_name) {
#if DEBUG
    DEBUGSTR << "find_variable(const string var_name)" << endl
	     << "\tParams: *" << var_name << "*" << endl;
#endif
    uint4 start = local_vars.size()-1;
    for (sint4 i=start; i>=0; --i)
      if (local_vars[i].first == var_name) 
	return var_location(VAR_LOCAL, local_vars[i].second);

    start = param_vars.size()-1;
    for (sint4 i=start; i>=0; --i)
      if (param_vars[i].first == var_name) { cout << "W " << param_vars[i].second << endl;
	return var_location(VAR_PARAM, param_vars[i].second); }

    // Look in current class for member variables
    start = curr_class->inst_data.size()-1;
    for (sint4 i=start; i>=0; --i)
      if (curr_class->inst_data[i].name == var_name)
	return var_location(VAR_INST_MEMBER, i);

    start = curr_class->static_data.size()-1;
    for (sint4 i = start; i>=0; --i)
      if (curr_class->static_data[i].name == var_name)
	return var_location(VAR_STATIC_MEMBER, i);

    return var_location(VAR_UNKNOWN, 0);
  }

  inline void Parser::add_block() {
#if DEBUG
    DEBUGSTR << "add_block()" << endl;
#endif
    local_scopes.push(local_vars.size());
  }

  inline void Parser::del_block() {
#if DEBUG
    DEBUGSTR << "del_block()" << endl;
#endif
    uint4 pop_to = local_scopes.top();
    local_scopes.pop();
    
    uint4 pops = local_vars.size() - pop_to;
    // fixme have to check
    cout << "pops " << pops << endl;
    FORU(i, pops) local_vars.pop_back();
    //    while (local_vars.size() != pop_to) {local_vars.pop_back(); cout << "0*" << endl;}

  }

  void Parser::parse() { 
#if DEBUG
    DEBUGSTR << "parse()" << endl;
#endif

    parse_declarations();

    token_stream_ptr = 0;
    parse_definitions();


    // Ensure that all classes are defined and functions defined
    // fixme add check
  }

  void Parser::parse_declarations() { 
#if DEBUG
    DEBUGSTR << "parse_declarations()" << endl;
#endif
    
    FOREVER {
      Token tk = get_token();
      TokenType type = tk.get_type();

      // Function definition outside class definition
      if (valid_type(type)) {
	Token tk2 = get_token();

	// Class member function
	if (tk2.get_type() == TK_ID) {
	  const string class_name = tk.get_value();
	  if (prog.class_ids.find(class_name) != prog.class_ids.end())
	    ERR("Unknown class");

	  curr_class = &prog.classes[prog.class_ids[class_name]];
	  
	  if (get_token().get_type() != TK_SCOPE)
	    ERR("Expecting '::'");

	  tk = get_token();
	  if (tk.get_type() != TK_ID)
	    PARSE_ERR("Expecting function name");

	  const string func_name = tk.get_value();
	  string stored_name = func_name;

	  if (get_token().get_type() != TK_LPAR)
	    ERR("Expecting '('");

	  // Get parameters
	  FOREVER {
	    tk = get_token();
	    if (tk.get_type() == TK_RPAR) break;
	    
	    TypeIndex t;
	    switch (tk.get_type()) {
	    case TK_BOOL: t = Type::TC_sint1; break;
	    case TK_BYTE: t = Type::TC_sint1; break;
	    case TK_UBYTE: t = Type::TC_uint1; break;
	    case TK_SHORT: t = Type::TC_sint2; break;
	    case TK_USHORT: t = Type::TC_uint2; break;
	    case TK_INT: t = Type::TC_sint4; break;
	    case TK_UINT: t = Type::TC_uint4; break;
	    case TK_LONG: t = Type::TC_sint8; break;
	    case TK_ULONG: t = Type::TC_uint8; break;
	    case TK_FLOAT: t = Type::TC_real4; break;
	    case TK_DOUBLE: t = Type::TC_real8; break;
	    case TK_STRING: t = Type::TC_string; break;
	    case TK_ID: {
	      string class_name = tk.get_value();
	      if (prog.class_ids.find(class_name) == prog.class_ids.end())
		prog.add_class(class_name);
	      
	      t = prog.classes[prog.class_ids[class_name]].ref_tind;
	      break;
	    }
	    default:
	      ERR("Unexpected token");
	    }
	    
	    stringstream s;
	    s << "*" << t;
	    stored_name += s.str();

	    tk = get_token();
	    if (tk.get_type() != TK_ID)
	      ERR("Expecting variable name for parameters");
	    
	    if (peek_token().get_type() == TK_COMMA) get_token();
	    else if (peek_token().get_type() != TK_RPAR)
	      ERR("Unexpected token");
	  }

	  if (curr_class->static_func_ids.find(stored_name) ==
	      curr_class->static_func_ids.end() &&
	      curr_class->virt_func_ids.find(stored_name) ==
	      curr_class->virt_func_ids.end())
	    ERR("Unknown function");	  
	}
	// Class constructor
	else if (tk2.get_type() == TK_SCOPE) {
	  const string class_name = tk.get_value();
	  if (prog.class_ids.find(class_name) != prog.class_ids.end())
	    ERR("Unknown class");

	  curr_class = &prog.classes[prog.class_ids[class_name]];

	  if (get_token().get_value() != class_name)
	    ERR("Constructor name not same as class name");

	  if (get_token().get_type() != TK_LPAR)
	    ERR("Expecting '('");

	  // fixme can i go something like a::a(a abc, int x) and
	  // have 'abc' be the 'this' being called?

	  string stored_name = class_name;

	  // Get parameters
	  FOREVER {
	    tk = get_token();
	    	    if (tk.get_type() == TK_RPAR) break;
	    
	    TypeIndex t;
	    switch (tk.get_type()) {
	    case TK_BOOL: t = Type::TC_sint1; break;
	    case TK_BYTE: t = Type::TC_sint1; break;
	    case TK_UBYTE: t = Type::TC_uint1; break;
	    case TK_SHORT: t = Type::TC_sint2; break;
	    case TK_USHORT: t = Type::TC_uint2; break;
	    case TK_INT: t = Type::TC_sint4; break;
	    case TK_UINT: t = Type::TC_uint4; break;
	    case TK_LONG: t = Type::TC_sint8; break;
	    case TK_ULONG: t = Type::TC_uint8; break;
	    case TK_FLOAT: t = Type::TC_real4; break;
	    case TK_DOUBLE: t = Type::TC_real8; break;
	    case TK_STRING: t = Type::TC_string; break;
	    case TK_ID: {
	      string class_name = tk.get_value();
	      if (prog.class_ids.find(class_name) == prog.class_ids.end())
		prog.add_class(class_name);
	      
	      t = prog.classes[prog.class_ids[class_name]].ref_tind;
	      break;
	    }
	    default:
	      ERR("Unexpected token");
	    }
	    
	    stringstream s;
	    s << "*" << t;
	    stored_name += s.str();

	    tk = get_token();
	    if (tk.get_type() != TK_ID)
	      ERR("Expecting variable name for parameters");
	    
	    if (peek_token().get_type() == TK_COMMA) get_token();
	    else if (peek_token().get_type() != TK_RPAR)
	      ERR("Unexpected token");
	  }
	  
	  // fixme check can constructor be static?
	  if (curr_class->virt_func_ids.find(stored_name) ==
	      curr_class->virt_func_ids.end())
	    ERR("Unknown constructor");
	}
	else ERR("Unexpected token");

	if (get_token().get_type() != TK_LBRACE)
	  ERR("Expecting '{'");

	// Skip over the function definition
	sint4 braces = 1;
	FOREVER {
	  tk = get_token();
	  if (tk.get_type() == TK_LBRACE) ++braces;
	  else if (tk.get_type() == TK_RBRACE) --braces;

	  if (braces == 0) break;
	}
      }
      else if (type == TK_CLASS) {
	tk = get_token();
	if (tk.get_type() != TK_ID)
	  ERR("expecting name for class");

	const string class_name = tk.get_value();

	if (prog.class_ids.find(class_name) == prog.class_ids.end())
	  curr_class = prog.add_class(class_name);
	else {
	  if (prog.classes[prog.class_ids[class_name]].defined)
	    ERR("class already used" + class_name);
	}
	tk = get_token();

	if (tk.get_type() == TK_COLON) {
	  tk = get_token();
	  if (tk.get_type() != TK_ID)
	    ERR("expecting name for superclass");

	  const string super_class = tk.get_value();

	  // Need to make sure super_class != derived class
	  if (class_name == super_class)
	    ERR("Derived class cannot be the same as the base class");
	  
	  if (prog.class_ids.find(super_class) == prog.class_ids.end())
	    ERR("Superclass does not exist");
	  
	  curr_class->super_class_index = prog.class_ids[super_class];
	  
	  // inherit
	  curr_class->inherit(prog.classes[curr_class->super_class_index]);

	  tk = get_token();	  
	}


	if (tk.get_type() != TK_LBRACE)
	  ERR("expecting class definition");

	parse_class_def();
	curr_class = 0;
      }
      else if (type == TK_EOF) return;
      else ERR("Unexpected token");
      
    }

  }

  
  void Parser::parse_class_def() {
#if DEBUG
    DEBUGSTR << "parse_class_def()" << endl;
#endif  
    // fixme in future check to see if defined already
    cout << curr_class << endl;
    if (curr_class->defined)
      ERR("Class already defined");
      //PARSE_ERR2("Class already defined", curr_class->name);
    curr_class->defined = true;
    curr_mod = PRIVATE;

    // Right now get public member variables and functions
    FOREVER {
      const Token type = get_token();
      switch(type.get_type()) {
      case TK_PUBLIC:   
	if (get_token().get_type() != TK_COLON)
	  ERR("Expecting ':' after modifier");
	curr_mod = PUBLIC;
	break;
      case TK_PROTECTED: 
	if (get_token().get_type() != TK_COLON)
	  ERR("Expecting ':' after modifier");
	curr_mod = PROTECTED;
	break;
      case TK_PRIVATE:
	if (get_token().get_type() != TK_COLON)
	  ERR("Expecting ':' after modifier");
	curr_mod = PRIVATE; 
	break;
      case TK_STATIC: 
	{
	  const Token type2 = get_token();
	  if (type2.get_type() == TK_VOID) parse_void_func(true);
	  else parse_var_or_array(type2, true);
	  break;
	}
      case TK_VOID:
	parse_void_func(false);
	break;
      case TK_SEMICOLON:
	break;
      case TK_RBRACE:
	return;
      default:
	parse_var_or_array(type, false);
      }
    }

  }

  void Parser::parse_void_func(bool is_static) {
#if DEBUG
    DEBUGSTR << "parse_void_func(bool is_static)" << endl
	     << "\tParams: *" << is_static << "*" << endl;
#endif
    const Token tk = get_token();
    
    // Check for conflicts
    if (tk.get_type() != TK_ID)
      ERR("Expecting function name");

    const string func_name = tk.get_value();

    parse_func_dec(TK_VOID, func_name, is_static, false);
  }

  void Parser::parse_var_or_array(const Token &type, const bool is_static) {
#if DEBUG
    DEBUGSTR << "parse_var_or_array(const Token &type, const bool is_static)" 
	     << endl << "\tParams: *" << type.get_value() << "*" 
	     << is_static << "*" << endl;
#endif
    const Token tk = peek_token();

    switch (type.get_type()) {
    case TK_BOOL:
    case TK_BYTE:
    case TK_UBYTE:
    case TK_SHORT:
    case TK_USHORT:
    case TK_INT:
    case TK_UINT:
    case TK_LONG:
    case TK_ULONG:
    case TK_FLOAT:
    case TK_DOUBLE:
    case TK_STRING:
      if (tk.get_type() == TK_LBRACKET) parse_array_dec(type, is_static);
      else parse_var_dec(type, is_static);
      return;
    case TK_ID:
      // look in classes
      if (prog.class_ids.find(type.get_value()) != 
	  prog.class_ids.end()) {
	if (tk.get_type() == TK_LBRACKET) parse_array_dec(type, is_static);
	else parse_var_dec(type, is_static);
      }
      else ERR("Unexpected token");
      return;
    default:
      ERR("Unexpected error");
    }

  }

  void Parser::parse_array_dec(const Token &type, const bool is_static) {
#if DEBUG
    DEBUGSTR << "parse_array_dec(const Token &type, const bool is_static)"
	     << endl << "\tParams: *" << type.get_value() << "*" 
	     << is_static << "*" << endl;
#endif
    get_token(); // the '['
    Token tk = get_token();
    if (tk.get_type() != TK_INTC && tk.get_type() != TK_UINTC)
      ERR("Expecting size");

    sint4 size = atoi(tk.get_value().c_str());

    if (size < 0) ERR("Array size needs to be bigger than 0");
    
    if (get_token().get_type() != TK_RBRACKET)
      ERR("Expecting ']'");

    // if there is a size...need to add a new array statement, otherwise
    // have to wait for array to be newed before it can be used

    // fixme need to see if name conflicts
    tk = get_token();
    const string name = tk.get_value();


    // fixme
#if 0
    DataMember dm;
    TypeIndex t;
    t.cat = Type::TC_array;
    switch (type.get_type()) {
    case TK_BOOL:   t.T_cat = Type::TC_sint1; break;
    case TK_BYTE:   t.T_cat = Type::TC_sint1; break;
    case TK_UBYTE:  t.T_cat = Type::TC_uint1; break;
    case TK_SHORT:  t.T_cat = Type::TC_sint2; break;
    case TK_USHORT: t.T_cat = Type::TC_uint2; break;
    case TK_INT:    t.T_cat = Type::TC_sint4; break;
    case TK_UINT:   t.T_cat = Type::TC_uint4; break;
    case TK_LONG:   t.T_cat = Type::TC_sint8; break;
    case TK_ULONG:  t.T_cat = Type::TC_uint8; break;
    case TK_FLOAT:  t.T_cat = Type::TC_real4; break;
    case TK_DOUBLE: t.T_cat = Type::TC_real8; break;
    case TK_STRING: t.T_cat = Type::TC_string; break;
    default: // fixme
      ;
    }
    dm.name = name;
    dm.type = t;
    if (is_static) { 
      dm.offset = curr_class->static_data.size();
      dm.is_static = true;
      curr_class->static_data.push_back(dm);
    }
    else {
      dm.offset = curr_class->inst_data.size();
      dm.offset = false;
      curr_class->inst_data.push_back(dm);
    }
#endif
  }




  void Parser::parse_var_dec(const Token &type, const bool is_static) {
#if DEBUG
    DEBUGSTR << "parse_var_dec(const Token &type, const bool is_static)"
	     << endl << "\tParams: *" << type.get_value() << "*" << is_static
	     << "*" << endl;
#endif

    if (type.get_value() == curr_class->name && 
	peek_token().get_type() == TK_LPAR)
      return parse_constructor(type);

    Token tk = get_token();
    if (tk.get_type() != TK_ID)
      ERR("Expecting name for variable");
    string name = tk.get_value();

    // function declaration
    if (peek_token().get_type() == TK_LPAR) {
      parse_func_dec(type, name, is_static, false);
      return;
    }

    FOREVER {
      DataMember dm;
      TypeIndex t;
      switch (type.get_type()) {
      case TK_BOOL:   t = Type::TC_sint1; break;
      case TK_BYTE:   t = Type::TC_sint1; break;
      case TK_UBYTE:  t = Type::TC_uint1; break;
      case TK_SHORT:  t = Type::TC_sint2; break;
      case TK_USHORT: t = Type::TC_uint2; break;
      case TK_INT:    t = Type::TC_sint4; break;
      case TK_UINT:   t = Type::TC_uint4; break;
      case TK_LONG:   t = Type::TC_sint8; break;
      case TK_ULONG:  t = Type::TC_uint8; break;
      case TK_FLOAT:  t = Type::TC_real4; break;
      case TK_DOUBLE: t = Type::TC_real8; break;
      case TK_STRING: t = Type::TC_string; break;

      default:
	ERR("FIXME");
      }

      dm.name = name;
      dm.type = t;
      dm.mod = curr_mod;
      if (is_static) { 
	dm.offset = curr_class->static_data.size();
	dm.is_static = true;
	curr_class->static_data_ids[dm.name] = dm.offset;
	curr_class->static_data.push_back(dm);
      } 
      else {
	dm.offset = curr_class->inst_data.size();
	dm.is_static = false;
	curr_class->inst_data_ids[dm.name] = dm.offset;
	curr_class->inst_data.push_back(dm);
      }

      cout << "NOW SIZE IS " << curr_class->inst_data.size() << endl;

      if (peek_token().get_type() == TK_ASSIGN) {
	// fixme
      }
	
      tk = peek_token();

      if (tk.get_type() == TK_COMMA) {
	get_token(); // the comma
	tk = get_token(); 
	if (tk.get_type() != TK_ID)
	  ERR("Expecting name for variable");
	name = tk.get_value();
      }
      else if (tk.get_type() == TK_SEMICOLON) {
	get_token(); // the semicolon
	return;
      }
      else return;
    }
      
  }


  void Parser::parse_constructor(const Token &type) {
#if DEBUG
    DEBUGSTR << "parse_constructor(const Token &type)" << endl
	     << "\tParams: *" << type.get_value() << "*" << endl;
#endif

    // Get paramaters
    // similar to parse_func_dec() ???

    parse_func_dec(type, type.get_value(), false, true);

      
  }


  void Parser::parse_func_dec(const Token &return_type, 
			      const string &func_name,
			      const bool is_static,
			      const bool is_constructor) {
#if DEBUG
    DEBUGSTR << "parse_func_dec(const Token &return_type, const string " 
	     << "&func_name, const bool is_static, const bool is_constructor)"
	     << endl
	     << "\tParams: *" << return_type.get_value() << "*" << func_name
	     << "*" << is_static << "*" << is_constructor << "*" << endl;
#endif

    // fixme
    FuncMember fm;
    fm.name = func_name;
    fm.mod = curr_mod;
    fm.is_static = is_static;
    // fixme: not always true
    //    fm.class_id = prog.classes.size()-1;
    fm.class_id = prog.class_ids[curr_class->name];


    // fixme void type...
    switch(return_type.get_type()) {
    case TK_BOOL: fm.ret_tind = Type::TC_sint1; break;
    case TK_BYTE: fm.ret_tind = Type::TC_sint1; break;
    case TK_UBYTE: fm.ret_tind = Type::TC_uint1; break;
    case TK_SHORT: fm.ret_tind = Type::TC_sint2; break;
    case TK_USHORT: fm.ret_tind = Type::TC_uint2; break;
    case TK_INT: fm.ret_tind = Type::TC_sint4; break;
    case TK_UINT: fm.ret_tind = Type::TC_uint4; break;
    case TK_LONG: fm.ret_tind = Type::TC_sint8; break;
    case TK_ULONG: fm.ret_tind = Type::TC_uint8; break;
    case TK_FLOAT: fm.ret_tind = Type::TC_real4; break;
    case TK_DOUBLE: fm.ret_tind = Type::TC_real8; break;
    case TK_STRING: fm.ret_tind = Type::TC_string; break;
    case TK_ID:
      if (prog.class_ids.find(return_type.get_value()) == prog.class_ids.end())
	PARSE_ERR("Unknown type");
      fm.ret_tind = prog.classes[prog.class_ids[return_type.get_value()]].ref_tind;
      break;
    default: // fixme
      ERR("fixme");
    }

    if (get_token().get_type() != TK_LPAR)
      PARSE_ERR("'(' expected after function name");


    // fixme this is wrong....
    // no check right now for 'this' pointer so it works
    // Push this pointer for non-static functions
    if (!is_static) {
      TypeIndex t;
      t = 0; // fixme need current function type
      fm.par_types.push_back(t);
    }

    string stored_name = fm.name;

    // Get parameters
    param_vars.clear();
    FOREVER {
      Token tk = get_token();
      if (tk.get_type() == TK_RPAR) break;

      TypeIndex t;
      switch (tk.get_type()) {
      case TK_BOOL: t = Type::TC_sint1; break;
      case TK_BYTE: t = Type::TC_sint1; break;
      case TK_UBYTE: t = Type::TC_uint1; break;
      case TK_SHORT: t = Type::TC_sint2; break;
      case TK_USHORT: t = Type::TC_uint2; break;
      case TK_INT: t = Type::TC_sint4; break;
      case TK_UINT: t = Type::TC_uint4; break;
      case TK_LONG: t = Type::TC_sint8; break;
      case TK_ULONG: t = Type::TC_uint8; break;
      case TK_FLOAT: t = Type::TC_real4; break;
      case TK_DOUBLE: t = Type::TC_real8; break;
      case TK_STRING: t = Type::TC_string; break;
      default: // fixme
	;
      }
      cout << "WHY " << t << endl;
      fm.par_types.push_back(t);
      
      stringstream s;
      s << "*" << t;
      stored_name += s.str();

      // fixme: For now assume var names have to be here
      tk = get_token();
      if (tk.get_type() != TK_ID)
	PARSE_ERR("Expecting variable name for parameters");

      param_vars.push_back(var_info(tk.get_value(), fm.par_types.size()-1));

      if (peek_token().get_type() == TK_COMMA) get_token();
      else if (peek_token().get_type() != TK_RPAR)
	PARSE_ERR2("Unexpected token in param list", peek_token().get_value());
    }

    // fixme need to check for conflicts.... and later on to see if fns defined

    // Set function into current class
    if (is_static) {
      cout << "Setting in func " << stored_name << " in " << prog.classes.size() << " " << curr_class->static_funcs.size() << endl;
      curr_class->static_func_ids[stored_name] = 
	curr_class->static_funcs.size();
      curr_class->static_funcs.push_back(fm);
      
    }
    else {
      cout << "Setting in func " << stored_name << " in " << prog.classes.size() << " " << curr_class->virt_funcs.size() << endl;
      curr_class->virt_func_ids[stored_name] = 
	curr_class->virt_funcs.size();
      curr_class->virt_funcs.push_back(fm);
    }

    cout << "WHAT IS MY SIZE " << curr_class->virt_funcs.size() << endl;

    Token tk = peek_token();
    if (tk.get_type() == TK_SEMICOLON) {
      get_token(); // the semicolon
      return;
    }
    else if (tk.get_type() == TK_LBRACE) {
      get_token(); // the '{'

      sint4 braces = 1;

      FOREVER {
	tk = get_token();
	if (tk.get_type() == TK_LBRACE) ++braces;
	else if (tk.get_type() == TK_RBRACE) --braces;
	
	if (braces == 0) break;
      }

      /*
      uint4 loc;
      if (is_static) {
	loc = curr_class->static_funcs.size() - 1;
	curr_func = &curr_class->static_funcs[loc];
      }
      else {
	loc = curr_class->virt_funcs.size() - 1;
	curr_func = &curr_class->virt_funcs[loc];
      }

      cout << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
      cout << curr_func->loc_types.size() << endl;
      parse_func_def();

      if (is_constructor && !has_return)
	curr_func->add_code(Opcodes::RETURN);

      */
      curr_func = 0;
    }
    else
      ERR("Unexpected token");
  }


  //////////////////////////////////////////////////////////////////
  // 2nd pass


  void Parser::parse_definitions() { 
#if DEBUG
    DEBUGSTR << "parse_definitions()" << endl;
#endif
    
    // A Class, class member function
    FOREVER {
      Token tk = get_token_info();
      TokenType type = tk.get_type();

      // fixme for class constructor
      if (valid_type(type)) {
	Token tk2 = get_token_info();
	if (tk2.get_type() == TK_ID) {
	  const string class_name = tk.get_value();
	  if (prog.class_ids.find(class_name) != prog.class_ids.end())
	    PARSE_ERR2("Unknown class", class_name);
	  
	  curr_class = &prog.classes[prog.class_ids[class_name]];
	  
	  // Should be the scope resolution operator
	  if (get_token_info().get_type() != TK_SCOPE)
	    PARSE_ERR("Expecting '::'");
	  
	  tk = get_token_info();
	  if (tk.get_type() != TK_ID)
	    PARSE_ERR("Expecting function name");

	  const string func_name = tk.get_value();
	  string stored_name = func_name;

	  if (get_token_info().get_type() != TK_LPAR)
	    PARSE_ERR("Expecting '('");
	  
	  // Get parameters
	  FOREVER {
	    tk = get_token_info();
	    if (tk.get_type() == TK_RPAR) break;
	    
	    TypeIndex t;
	    switch (tk.get_type()) {
	    case TK_BOOL: t = Type::TC_sint1; break;
	    case TK_BYTE: t = Type::TC_sint1; break;
	    case TK_UBYTE: t = Type::TC_uint1; break;
	    case TK_SHORT: t = Type::TC_sint2; break;
	    case TK_USHORT: t = Type::TC_uint2; break;
	    case TK_INT: t = Type::TC_sint4; break;
	    case TK_UINT: t = Type::TC_uint4; break;
	    case TK_LONG: t = Type::TC_sint8; break;
	    case TK_ULONG: t = Type::TC_uint8; break;
	    case TK_FLOAT: t = Type::TC_real4; break;
	    case TK_DOUBLE: t = Type::TC_real8; break;
	    case TK_STRING: t = Type::TC_string; break;
	    case TK_ID: {
	      string class_name = tk.get_value();
	      if (prog.class_ids.find(class_name) == prog.class_ids.end())
		prog.add_class(class_name);
	      
	      t = prog.classes[prog.class_ids[class_name]].ref_tind;
	      break;
	    }
	    default:
	      PARSE_ERR("Unexpected token");
	    }
	    
	    stringstream s;
	    s << "*" << t;
	    stored_name += s.str();

	    tk = get_token_info();
	    if (tk.get_type() != TK_ID)
	      PARSE_ERR("Expecting variable name for parameters");
	    
	    if (peek_token_info().get_type() == TK_COMMA) get_token_info();
	    else if (peek_token_info().get_type() != TK_RPAR)
	      PARSE_ERR("Unexpected token");
	  }
	  if (curr_class->static_func_ids.find(stored_name) !=
	      curr_class->static_func_ids.end())
	    curr_func = &curr_class->static_funcs[curr_class->static_func_ids[stored_name]];
	  else if (curr_class->virt_func_ids.find(stored_name) !=
		   curr_class->virt_func_ids.end())
	    curr_func = &curr_class->virt_funcs[curr_class->virt_func_ids[stored_name]];
	  else PROGRAM_ERR;
	}
	// Class constructor
	else if (tk2.get_type() == TK_SCOPE) {
	  	  const string class_name = tk.get_value();
	  if (prog.class_ids.find(class_name) != prog.class_ids.end())
	    ERR("Unknown class");

	  curr_class = &prog.classes[prog.class_ids[class_name]];

	  if (get_token_info().get_value() != class_name)
	    PARSE_ERR("Constructor name not same as class name");

	  if (get_token_info().get_type() != TK_LPAR)
	    PARSE_ERR("Expecting '('");

	  // fixme can i go something like a::a(a abc, int x) and
	  // have 'abc' be the 'this' being called?

	  string stored_name = class_name;

	  // Get parameters
	  FOREVER {
	    tk = get_token_info();
	    if (tk.get_type() == TK_RPAR) break;
	    
	    TypeIndex t;
	    switch (tk.get_type()) {
	    case TK_BOOL: t = Type::TC_sint1; break;
	    case TK_BYTE: t = Type::TC_sint1; break;
	    case TK_UBYTE: t = Type::TC_uint1; break;
	    case TK_SHORT: t = Type::TC_sint2; break;
	    case TK_USHORT: t = Type::TC_uint2; break;
	    case TK_INT: t = Type::TC_sint4; break;
	    case TK_UINT: t = Type::TC_uint4; break;
	    case TK_LONG: t = Type::TC_sint8; break;
	    case TK_ULONG: t = Type::TC_uint8; break;
	    case TK_FLOAT: t = Type::TC_real4; break;
	    case TK_DOUBLE: t = Type::TC_real8; break;
	    case TK_STRING: t = Type::TC_string; break;
	    case TK_ID: {
	      string class_name = tk.get_value();
	      if (prog.class_ids.find(class_name) == prog.class_ids.end())
		prog.add_class(class_name);
	      
	      t = prog.classes[prog.class_ids[class_name]].ref_tind;
	      break;
	    }
	    default:
	      PARSE_ERR("Unexpected token");
	    }
	    
	    stringstream s;
	    s << "*" << t;
	    stored_name += s.str();

	    tk = get_token_info();
	    if (tk.get_type() != TK_ID)
	      PARSE_ERR("Expecting variable name for parameters");
	    
	    if (peek_token_info().get_type() == TK_COMMA) get_token_info();
	    else if (peek_token_info().get_type() != TK_RPAR)
	      PARSE_ERR("Unexpected token");
	  }
	  
	  // fixme check can constructor be static?
	  if (curr_class->virt_func_ids.find(stored_name) !=
	      curr_class->virt_func_ids.end())
	    curr_func = &curr_class->virt_funcs[curr_class->virt_func_ids[stored_name]];
	  else PROGRAM_ERR;
	  
	  
	}
	else
	  PARSE_ERR("Expecting class name for function definition");

	// fixme get the f