
#include "SimplePathfinder.H"
//#include "SimpleGlobalPF.H"
//---------------------------------------------------
SimplePathfinder::SimplePathfinder(GlobalPathfinder &gp_, PathExecutor &pe_)
  : TerrainBasicImp<PathfindTask>(), pe(pe_), gp(gp_), initiated(false)
{ 
}
//---------------------------------------------------
SimplePathfinder::SimplePathfinder(const SimplePathfinder &other) 
  : TerrainBasicImp<PathfindTask>(other), pe(other.pe), gp(other.gp), initiated(other.initiated)
{
}


//---------------------------------------------------
void SimplePathfinder::add_obj(const Object *obj){
  assert(initiated);
  TerrainBasicImp<PathfindTask>::add_obj(obj);
  pe.add_obj(obj);
  gp.insert_object(obj);
}

//---------------------------------------------------
void SimplePathfinder::update_obj(const Object *obj){
  assert(initiated);
  TerrainBasicImp<PathfindTask>::update_obj(obj);
  pe.update_obj(obj);
  gp.update_object(obj);
}

//---------------------------------------------------
void SimplePathfinder::remove_obj(const Object *obj){
  assert(initiated);
  // shorthand
  typedef ObjIsGoalInTaskId::iterator OGIT;

  // If this is the target of any existing task...
  std::pair<OGIT,OGIT> range = objIsGoalInTaskId.equal_range(obj);
  for( OGIT i = range.first ; i != range.second ; i++ ) {
    // Change the target of the task to the last known location of the object
    FIND(taskId2task, ti2tIt, i->second);
    assert(ti2tIt != taskId2task.end());
    ti2tIt->second->task.goal.target = Goal::LOCATION;
    ti2tIt->second->task.goal.obj->get_center(ti2tIt->second->task.goal.loc.x, 
                                              ti2tIt->second->task.goal.loc.y);
  }
  
  TerrainBasicImp<PathfindTask>::remove_obj(obj);
  pe.remove_obj(obj);
  gp.remove_object(obj);
}

//---------------------------------------------------
void SimplePathfinder::add_segments(const Vector<Segment> &segs){
  assert(initiated);
  TerrainBasicImp<PathfindTask>::add_segments(segs);
  pe.add_segments(segs);
  gp.insert_boundaries(segs);
  invalidate_paths(segs);
}

//---------------------------------------------------
void SimplePathfinder::cancel_task(const Object *obj)
{
  TerrainBasicImp<PathfindTask>::cancel_task(obj);
}

//---------------------------------------------------
void SimplePathfinder::remove_task(TaskId tid)
{
  TerrainBasicImp<PathfindTask>::remove_task(tid);
}

//--------------------------------------------------

real8 SimplePathfinder::find_path(const Object* /*obj*/, const Goal& /*goal*/, Path* /*path*/, ConsiderObjects /*consider*/)
{ 
  assert(0);
  return -1;
//   assert( goal.mode == Goal::TOUCH && "ATTACK, VICINITY modes are currently unimplemented");
//   assert( (goal.target == Goal::LOCATION || goal.target == Goal::OBJ) && 
// 	  "Unknown Goal::target type");
  
//   Path result;
//   bool found;
//   if (goal.target == Goal::LOCATION) {
//     found = pfEngine->find_path(obj, goal.loc, result,  consider);
//   } else {
//     found = pfEngine->find_path(obj, goal.obj, result,  consider);
//   }
  
//   if( !found )
//     return -1;
  
//   // compute and return the length of the path
//   return path_length_from(result, get_center_loc(obj));
  
}
//--------------------------------------------------

real8 SimplePathfinder::find_path(const Loc &/*l1*/, const Loc &/*l2*/, sint4 /*radius*/, Path * /*path*/, ConsiderObjects /*consider*/)
{
  assert(0); // unimplemented yet    
  return -1;
}

//---------------------------------------------------
void SimplePathfinder::execute_tasks(Vector<MoveCmd> &cmds, Vector<StatusMsg> &msgs){
  assert(initiated);
  //tell the path execution system to execute the tasks
  pe.execute_tasks(tasks, cmds, msgs);
  
  //Remove all tasks who have had messages sent about them
  //It has been suggested that maybe the tasks shouldn't be removed if
  //the failure is a movement failure, but for now, that's the way it is...
  FORALL (msgs, t) {
   clean_task(t->task_id);
  }
}
//---------------------------------------------------

void SimplePathfinder::init(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 client_pID_, sint4 neutral_pID_)
{
 
  TerrainBasicImp<PathfindTask>::init(tiles_x_, tiles_y_, tile_points_, client_pID_, neutral_pID_);
  gp.init(tiles_x_, tiles_y_, tile_points_, client_pID_, neutral_pID_);

//   Vector<sint4> levels_of_detail;
//   levels_of_detail.push_back(4);  // default - use a resolution reduced by a factor of 4.
//   levels_of_detail.push_back(1);  // also maintain a full resolution view.
//   gp = boost::shared_ptr<GlobalPathfinder>(new SimpleTerrain::SimpleGlobalPF(tiles_x_, tiles_y_, tile_points_, levels_of_detail));
  initiated = true;
}
//---------------------------------------------------
void SimplePathfinder::invalidate_paths(const Vector<Segment> &segments)
{
  FORALL(tasks, ti) {
    const Object *obj = *(ti->task.objs.begin());
    // fixme: better handling of zcat != ON_LAND ?
    if (obj->get_zcat() != Object::ON_LAND ) continue;

    FORALL(segments, si) {
      if (is_intersect(*si, ti->path) ) {
	// this path is invalid, mark as pending for (re)planning
	pendingTasks.push_back(ti);
	ti->isPending = true;
	std::cout << "Marked an invalid path" << std::endl;
	break;
      }
    }
  }
}

bool SimplePathfinder::plan_tasks(uint4 max_time)
{
  assert(initiated);
  sint4 stop_time = 0;
  if(max_time > 0){
    stop_time = (sint4)SDL_GetTicks() + max_time;
  }

  // cout << "Chance To Plan Task\n" ;

  //Continue planning until we have reached the max_time or we aren't planning anymore
  while( (sint4)SDL_GetTicks() < stop_time && plan_pending_task()){
  }

  //returns true if the full alotment of time was taken
  return !((sint4)SDL_GetTicks() < stop_time);


}

//----------------------------------------------------------------------------------
bool SimplePathfinder::plan_pending_task()
{
  //if pendingTasks is empty, there are no more tasks to plan
  if (pendingTasks.empty()) return false;

  TIter tit = pendingTasks.front(); pendingTasks.pop_front();
  // mark the task as handled
  tit->isPending = false;

  assert(!tit->task.objs.empty());

  // plan
  plan_task(tit);
  return true;
}

void SimplePathfinder::plan_task(TIter tit)
{
  Path newpath;
  bool found;

  found = gp.find_path(tit->task, newpath, IGNORE_ALL_MOBILE_OBJS);

  if( found ) {
    if (tit->task.goal.target == Goal::LOCATION) {
      newpath.locs[newpath.locs.size()-1] = tit->task.goal.loc; //to make up for a low resolution map, the last location is replaced with the exact correct location
    }
    assert(newpath.locs.size() > 0);
    tit->isPending = false;
    tit->path = newpath;
    tit->current_wp = 0; //set to 1 instead of 0, so as to skip the first wp (which should be the current location)
  }
  else {
    tit->noPathFound = true;
  }
}


real8 SimplePathfinder::path_length(const Path &path)
{
  if( path.locs.size() < 2 )
    return 0;
  real8 total = 0;
  typeof(path.locs.begin()) i = ++path.locs.begin();
  for(; i != path.locs.end() ; i++ )
    total += i->distance(*(i-1));
  return total;
}

TerrainBase::Loc SimplePathfinder::get_center_loc(const Object *obj)
{
  Loc l;
  obj->get_center(l.x, l.y);
  return l;
}

bool SimplePathfinder::is_intersect(const Segment &segment, const Path &path)
{
  typeof(path.locs.begin()) pj = path.locs.begin();
  typeof(path.locs.begin()) pi = pj++;
  for( ; pj != path.locs.end() ; pi++, pj++ ) {
    if (Segment(*pi, *pj).touches(segment) )
      return true;
  }
  return false;
}
