#ifndef ST_FORCEFIELD_H
#define ST_FORCEFIELD_H

//Comments coming soon, sorry...
// This is an ORTS file
// (c) Michael Buro, Sami Wagiaalla, David Deutscher, Nick Wiebe
// licensed under the GPL

// Local Pathfinder using ForceFields to execute paths,
// and interface to global (simple a*) pathfinder.


#include "TerrainBasicImp.H"
#include "Random.H"
#include "Coord.H"
#include "SimpleTerrain.H"

#include "boost/shared_ptr.hpp"

namespace SimpleTerrain{

  typedef TerrainBase::Loc      Loc;
  typedef TerrainBase::Segment  Segment;
  typedef TerrainBase::Task     Task;

  /** Structure to keep data associated to a path
      fixme: right now this inherits from UnitPath, from SimpleTerrain,
      but that doesn't make much sense. This was done primarily for convenience.
   */
  struct UnitPath_FF : public UnitPath
  {
    /// Path data
    //@{
    //const Object *obj;
    //Vector<Loc> path;
    //Loc goal;
    //int current_target; // index of current way point in path vector
    //bool no_orders_yet; // true at the beginning, before the first move command is issued
    //@}

    /// Collision avoidance and resolution
    //@{
    // If adding fields - check ST_Task::set_path for needed updates.
    //bool stopped_to_avoid_collision; // true if the object was stopped temporarily to avoid a collision
    //int collision_step; // this is incremented while solving unexpected stops
    //uint4 wakeup;       // if SDL_GetTicks() >= wakeup then this UnitPath is not
                        // sleeping ie not to be skipped over.
    //Loc spot;           // spot of the last unexpected collision
    //@}

    bool replanned;

    UnitPath_FF(const Object *obj_, const Vector<Loc> &path_) : UnitPath(obj_, path_), replanned(false){};
    ~UnitPath_FF() {};

    //void mark_spot(const Loc &p) { spot = p; };
    //void sleep(int time) { wakeup = SDL_GetTicks() + time; };
    //bool sleeping() const { return wakeup > SDL_GetTicks(); };

    //bool is_intersect(const Segment &) const;
  };


  /** Functor that allows Locations to be put into sets*/
  class Loc_less : public std::binary_function<Loc, Loc, bool>{
  public:
    bool operator() (Loc a, Loc b) const;
  };

  /** Two Dimensional Vector */
  class Vec_2D{
  public:
    sint4 x;/// x-coord
    sint4 y;/// y-coord
    Vec_2D(sint4 x_=0, sint4 y_=0) :  x(x_), y(y_){};
    Vec_2D(const Vec_2D &other) :  x(other.x), y(other.y){};
    Vec_2D(const Loc &a, const Loc &b) : x(b.x - a.x), y(b.y - a.y) {};
    Vec_2D operator+(const Vec_2D &other) const; ///Vector addition
    Vec_2D operator-(const Vec_2D &other) const; ///Vector subtraction
    Vec_2D project_onto(const Vec_2D &other) const; /// Project the vector onto the line defined by the given vector
    sint4 dot_prod(const Vec_2D &other) const;
    real8 length() const;
    Vec_2D multiply(real8 k) const;
    bool operator==(const Vec_2D &other) const;
    bool operator!= (const Vec_2D &other) const;

    inline static sint4 round(real8 real) {
      return (sint4)floor(real + 0.5);
    }
  };

  /// The internal data complementing Task data:
  struct FF_Task {
    Task task;
    bool isPending;
    bool isFailed;
    bool noPathFound;
    boost::shared_ptr<UnitPath_FF> path;    // todo: replace with a cleaner implementation
    //Vector<const Object*> followers;
    uint4 replanned;
    //sint4 ticks; //used for automatic replanning when moving to a mobile object

    FF_Task(const Task& task_)
      : task(task_), isPending(true), isFailed(false), noPathFound(false), replanned(0) {};
    FF_Task(const FF_Task& other)
      : task(other.task), isPending(other.isPending),
	isFailed(other.isFailed), noPathFound(other.noPathFound),
      path(other.path), /*followers(other.followers),*/ replanned(other.replanned) {};
    void set_path(const Object *obj_, const Vector<Loc> &path_);
    //void init_path(const Object *obj_, const Vector<Loc> &path_) { path=boost::shared_ptr<UnitPath>(new UnitPath(obj_, path_)); };
    bool isValid() const { return !isPending && !isFailed && !noPathFound && path; };
  private:
    FF_Task operator=