#ifndef PATHEXECUTOR_H
#define PATHEXECUTOR_H

#include "TerrainBase.H"
#include "PathfindTask.H"
/** 
  The PathExecutor is responsible for translating a path and a pathfinding task
  into the individual move commands for the units. It takes a task and a path for the
  task, and reports when the task is completed as well as if it was completed 
  successfully or not.
*/
class PathExecutor{
  
public:

  PathExecutor(){};
  virtual ~PathExecutor(){};

  /** Generate all the necessary move commands and Status messages (success, movement failure, no path failure)*/
  virtual void execute_tasks(std::list<PathfindTask> &tasks, Vector<TerrainBase::MoveCmd> &cmds, Vector<TerrainBase::StatusMsg> &msgs) = 0;
 
  /**
     \name View Msg Methods
     If a PathExecutor needs to keep track of objects or segments, then
     it can redefine these virtual functions 
   */
  //@{
  virtual void add_obj(const Object * /*obj*/){};
  virtual void update_obj(const Object * /*obj*/){};
  virtual void remove_obj(const Object* /*obj*/){};
  virtual void add_segments(const Vector<TerrainBase::Segment>& /*s*/){};
  //@}
};

#endif
