/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Global.H"
#include "RT_Client.H"
#include "RT_ClientEnv.H"
#include "RTS_Options.H"
#include "CompActStd.H"
#include "CompActExp1.H"
#include "CompActRndMove.H"

RT_Client::~RT_Client() {}

void RT_Client::run()
{
  env.local = true; // generate vics for local actions

  string cpact; // compact
  env.opt.get("-cpact", cpact);
  
  auto_ptr<CompAction> compact;

  // casts make gcc-3.1.1 happy
  if      (cpact == "std")
    compact = static_cast<auto_ptr<CompAction> >(new CompActStd);
  else if (cpact == "exp1")
    compact = static_cast<auto_ptr<CompAction> >(new CompActExp1);
  else if (cpact == "rnd")
    compact = static_cast<auto_ptr<CompAction> >(new CompActRndMove);
  else {
    cerr << cpact << endl;
    ERR("unknown compact");
  }

  FOREVER {

    RT_ClientEnv::Event ev = env.get_event();

    if (ev != RT_ClientEnv::EVT_NONE) {
      
      if (ev == RT_ClientEnv::EVT_STOP) return;

      if (ev == RT_ClientEnv::EVT_NEW_OBJS) {

	// compute actions

	FORU (i, env.vics.size()) {
	  Action act = compact->action(env.vics[i]);
	  env.set_action(env.vics[i][0].sv_id, act);
	}

#if 0
	// timout test
	if (cpact == "std") {
	  sint4 s = random() % 3 + 0;
	  cout << "SLEEP " << s << endl;
	  sleep(s);
	}
#endif
	
	// cout << "actions computed" << endl;
#if 0	
	usleep(150000);
#endif	
	env.send_actions();
      }
    }
  }
}
